
#include "error.h"
#include "sound.h"

#ifdef HAVE_ALSA

int get_playback_volume_all(snd_mixer_elem_t *elem){
    int chn, n = 0;
    long int sum = 0;

    for (chn = 0; chn <= SND_MIXER_SCHN_LAST; chn++){
        if (!snd_mixer_selem_has_playback_channel(elem, chn)) continue;
        //snd_mixer_selem_get_playback_volume(elem, 
        n++;
    }
    return sum/n;

}


void alsa_pcm_info(char *pcm_name, enum sound_mode mode){
    //int err;
   // snd_pcm_hw_params_t *hwparams;
//    snd_pcm_t *pcm;
/*    snd_pcm_format_t formats[] = {SND_PCM_FORMAT_MU_LAW, SND_PCM_FORMAT_U8, SND_PCM_FORMAT_S16};
    char *formats_str[] = {"MULAW", "U8", "S16"};*/
	enum sound_frame_fmt formats[] = {SOUND_FMT_U8, SOUND_FMT_S16};
	
    int rates[] = {8000, 16000, 44100, 48000};
    int f, c, r;

/*    err=snd_pcm_open(&pcm, pcm_name, dirflg, 0);
    if (err<0) return;

    snd_pcm_hw_params_alloca(&hwparams);
    snd_pcm_close(pcm); */

    init_debug(-1, NULL);
    printf("        modes: ");
    for (f = 0; f < 2; f++){
        for (c = 1; c <= 2; c++){
            for (r = 0; r < 4; r++){
                struct sound *sound;
				struct sound_stream *stream;

                sound = init_sound(mode);
                if (!sound) continue;
                
                if (mode == SOUND_PLAY)
					stream = sound->play;
				else
					stream = sound->rec;

				stream->datafmt->fmt = formats[f];
				stream->datafmt->channels = c;
				stream->datafmt->rate = rates[r];
				stream->datafmt->per_time_us =  50000;
				stream->datafmt->buf_time_us = 100000;
                
                if (sound_stream_open(stream, pcm_name) < 0) continue;
				
                printf("%s_%d_%d  ", sound_frame_fmt2str(formats[f]), 
						c, rates[r]);

                sound_stream_close(stream);
                free_sound(sound);
            }
        }
    }
    printf("\n");
}

        

void alsa_mixer_info(char *mixer_name, snd_pcm_stream_t dir){
    snd_mixer_t *mixer_handle;
    snd_mixer_elem_t *elem;
    snd_ctl_t *ctl_handle;
    snd_ctl_card_info_t *hw_info;
    int sw;
    const char *mixer_chip;
    int err;

    snd_ctl_card_info_alloca(&hw_info);

    err = snd_ctl_open(&ctl_handle, mixer_name, 0);
    if (err<0) goto x3c;
    err = snd_ctl_card_info(ctl_handle, hw_info);
    if (err<0){
        snd_ctl_close(ctl_handle);
        goto x3c;
    }
    mixer_chip=snd_ctl_card_info_get_mixername(hw_info);
    printf("        mixer chip=%s\n", mixer_chip);
    snd_ctl_close(ctl_handle);

x3c:        
    err=snd_mixer_open(&mixer_handle, 0);
    if (err<0) goto x3;
    err=snd_mixer_attach(mixer_handle, mixer_name);
    if (err<0) goto x3;
    err=snd_mixer_selem_register(mixer_handle, NULL, NULL);
    if (err<0) goto x3;
    err=snd_mixer_load(mixer_handle);
    if (err<0) goto x3;

//        printf("        elements=");
    for (elem=snd_mixer_first_elem(mixer_handle);
         elem;
         elem=snd_mixer_elem_next(elem)){

        printf("        %-17s", snd_mixer_selem_get_name(elem));


        if (snd_mixer_selem_has_playback_volume(elem) && dir == SND_PCM_STREAM_PLAYBACK){
            printf("\tplay: ");
            if (snd_mixer_selem_is_playback_mono(elem))
                printf("mo ");
            else
                printf("st ");

        }

        if (snd_mixer_selem_has_capture_volume(elem) && dir == SND_PCM_STREAM_CAPTURE){
            printf("\tcapture: ");
            if (snd_mixer_selem_is_capture_mono(elem))
                printf("mo ");
            else
                printf("st ");
//                snd_mixer_selem_get_playback_volume(
        }

        if (snd_mixer_selem_has_capture_switch(elem)) {
            printf("rec");
              
            err=snd_mixer_selem_get_capture_switch(elem, 0, &sw);
            if (err<0) goto x3;
            
            if (sw) printf("src");
        }
        printf("\n");
    }
    printf("\n");

x3:
    snd_mixer_close(mixer_handle);
    
    
}


void card_info(int card){
    int pcm_device;
    char *name;
    snd_ctl_t *ctl;
    char dev[256], s[256];
    snd_pcm_info_t *pcm_info;

    printf("card hw:%d ",card);

    if (snd_card_get_name(card, &name)==0){
        printf("%s", name);
    }
    printf("\n");                

    sprintf(dev,"hw:%d", card);
    snd_ctl_open(&ctl, dev, 0);
    
    snd_pcm_info_alloca(&pcm_info);    

    pcm_device=-1;
    while(1){
        if (snd_ctl_pcm_next_device(ctl, &pcm_device)) pcm_device=-1;
        if (pcm_device<0) break;

        snd_pcm_info_set_device(pcm_info, pcm_device);
        snd_pcm_info_set_subdevice(pcm_info, 0);
        snd_pcm_info_set_stream(pcm_info, SND_PCM_STREAM_PLAYBACK);
        
        if (snd_ctl_pcm_info(ctl, pcm_info)<0) continue;
        
        printf("    playback hw:%d,%d %s\n", card, pcm_device, snd_pcm_info_get_name(pcm_info)); 
        sprintf(s, "hw:%d,%d", card, pcm_device);
        alsa_pcm_info(s, SOUND_PLAY);
    }
                        
    pcm_device=-1;
    while(1){
        if (snd_ctl_pcm_next_device(ctl, &pcm_device)) pcm_device=-1;
        if (pcm_device<0) break;

        snd_pcm_info_set_device(pcm_info, pcm_device);
        snd_pcm_info_set_subdevice(pcm_info, 0);
        snd_pcm_info_set_stream(pcm_info, SND_PCM_STREAM_CAPTURE);
        
        if (snd_ctl_pcm_info(ctl, pcm_info)<0) continue;
        
        printf("    capture  hw:%d,%d %s\n", card, pcm_device, snd_pcm_info_get_name(pcm_info)); 
        sprintf(s, "hw:%d,%d", card, pcm_device);
        alsa_pcm_info(s, SOUND_REC);
    }
}

void alsa_info(void){
    int card;
    
    printf("\n  alsa_info:\n");
    card=-1;
    snd_card_next(&card);
    while(card>-1){
        card_info(card);
        snd_card_next(&card);
    }
    
    printf("\n");
}

#endif
