
#include "config.h"
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"

#ifdef HAVE_ALSA
#include <alsa/asoundlib.h>
#endif

void do_not_optimize_here(void *p)
{
    /* stop GCC optimization - avoid bugs in it */
}

static FILE *debug_file=NULL;
static int debug_quiet = 0;

void init_debug(int debug_type, char *filename){
    switch (debug_type){
        case -1:
            if (debug_file != NULL && debug_file != stderr) fclose(debug_file);
            debug_file=NULL;
            debug_quiet=1;
            break;
        case 0:
            if (debug_file != stderr) fclose(debug_file);
            debug_file=NULL;
            debug_quiet=0;
            break;
        case 1:
            debug_file=fopen(filename, "at");
            debug_quiet=0;
            break;
        case 2:
            debug_file=stderr;
            debug_quiet=0;
            break;    
    }
}

void free_debug(){
    if (!debug_file) return;
    if (debug_file!=stderr) fclose(debug_file);
    debug_file = NULL;
}

static inline void force_dump(void){
    fprintf(stderr, "Forcing core dump");
    fflush(stderr);
    raise(SIGSEGV);
    exit(1);
}

void debug_internal(char *file, int line, char *m, ...){
    va_list l;
    
    va_start(l, m);
    fprintf(stderr, "INTERNAL ERROR at %s:%d ", file, line);
    vfprintf(stderr, m, l);
    fprintf(stderr, "\n");
    force_dump();
}

void dbg(char *m, ...){
    va_list l;

    if (!debug_file) return;

    va_start(l, m);
    vfprintf(stderr, m, l);
    va_end(l);
    fflush(stderr);
} 

void error(char *m, ...){
    va_list l;

    if (debug_quiet) return;
    va_start(l, m);
    vfprintf(stderr, m, l);
    va_end(l);
    fflush(stderr);
} 


void errormsg(char *m, ...){
    va_list l;
	int err;
 	char errbuf[256];

    if (debug_quiet) return;
    err = errno;
    va_start(l, m);
    vfprintf(stderr, m, l);
    va_end(l);
     
	fprintf(stderr, " - Error %d %s\n", err, STRERR(err));
    fflush(stderr);
} 

#if HAVE_ALSA
void snderrormsg(int snderr, char *m, ...){
    va_list l;

    if (debug_quiet) return;
    va_start(l, m);
    vfprintf(stderr, m, l);
    va_end(l);
     
	fprintf(stderr, " - Error %d %s\n", snderr, snd_strerror(-snderr));
    fflush(stderr);
} 
#endif


#ifdef HAVE_LIBFTDI
void ftdierrormsg(struct ftdi_context *ftdi, char *m, ...){
    va_list l;

    if (debug_quiet) return;
    va_start(l, m);
    vfprintf(stderr, m, l);
    va_end(l);
     
	fprintf(stderr, " - Error %s\n", ftdi_get_error_string(ftdi));
    fflush(stderr);
}
#endif
