#ifndef __ERROR_H
#define __ERROR_H

#include <errno.h>

#ifdef HAVE_LIBFTDI_STATIC
#include "../libftdi/ftdi.h"
#else
#ifdef HAVE_LIBFTDI
#include <ftdi.h>
#endif
#endif

#define internal(args...) debug_internal(__FILE__, __LINE__, args)


#if defined __USE_XOPEN2K && !defined __USE_GNU
// #warning POSIX version returns int
#define STRERR(no) (strerror_r(no, errbuf, sizeof(errbuf)), errbuf)
#else
// #warning GNU version returns char*
#define STRERR(no) strerror_r(no, errbuf, sizeof(errbuf))
#endif


#define SINSTR(sa) (snprintf(sinbuf, sizeof(sinbuf), "%d.%d.%d.%d:%d", \
        sa.sin_addr.s_addr & 0xff, \
        (sa.sin_addr.s_addr >> 8) & 0xff, \
        (sa.sin_addr.s_addr >> 16) & 0xff, \
        (sa.sin_addr.s_addr >> 24) & 0xff, \
        ntohs(sa.sin_port)), sinbuf)


void init_debug(int debug_type, char *filename);
void free_debug(void);

void debug_internal(char *file, int line, char *s, ...);

void dbg(char *m, ...);
void error(char *m, ...);
void errormsg(char *m, ...); // without trailing \n !

#ifdef HAVE_ALSA
void snderrormsg(int snderr, char *m, ...);
#endif

#ifdef HAVE_LIBFTDI
void ftdierrormsg(struct ftdi_context *ftdi, char *m, ...);
#endif

#endif
