#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

#include "sound.h"
#include "error.h"
#include "rc.h"

//static const char *clabels[] = SOUND_DEVICE_LABELS;
static const char *cnames[] = SOUND_DEVICE_NAMES;

void oss_mixer_info(void){
    int k;
    char filename[64];
    
    printf("\n  oss_mixer_info\n");
    
    oss_mixer_file_info("/dev/mixer");
    
    for (k=0; k<16; k++){
        sprintf(filename, "/dev/mixer%d", k);
        oss_mixer_file_info(filename);
    }
    
    oss_mixer_file_info("/dev/sound/mixer");
    
    for (k=0; k<16; k++){
        sprintf(filename, "/dev/sound/mixer%d", k);
        oss_mixer_file_info(filename);
    }
}

int oss_mixer_file_info(char * filename){
    int ret, i, fd, devmask, recmask, recsrc, stereodevs, volume;
    struct stat st;

    ret = lstat(filename, &st);
    if (ret < 0) return -1;
    if (!S_ISCHR(st.st_mode)) return -1;
    
    fd = open(filename, O_RDWR, 0);
    if (fd < 0) return -1;
  
    printf("%s\n", filename);
    if (ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devmask) < 0){
        errormsg("Failed ioctl(SOUND_MIXER_READ_DEVMASK)");
        return 1;
    }
    
    if (ioctl(fd, SOUND_MIXER_READ_RECMASK, &recmask) < 0){
        errormsg("Failed ioctl(SOUND_MIXER_READ_RECMASK)");
        return 1;
    }

    if (ioctl(fd, SOUND_MIXER_READ_RECSRC, &recsrc) < 0){
        errormsg("Failed ioctl(SOUND_MIXER_READ_RECSRC)");
        return 1;
    }

    if (ioctl(fd, SOUND_MIXER_READ_STEREODEVS, &stereodevs) < 0){
        errormsg("Failed ioctl(SOUND_MIXER_READ_STEREODEVS)");
        return 1;
    }

    printf("channel     left right\n");
    for (i=0; i<SOUND_MIXER_NRDEVICES; i++){
        if ((devmask & (1<<i))==0) continue;

        printf("%-8s ", cnames[i]); 
        
        if (ioctl(fd, MIXER_READ(i), &volume) < 0){
            errormsg("Failed ioctl(MIXER_READ(%d))", i);
            goto next;
        }

        if (stereodevs & (1<<i)) printf("st ");
        else  printf("mo ");
        printf(" %3d %3d ", volume & 0xff, volume>>8);

        if (recmask & (1<<i)) printf("rec");
        if (recsrc & (1<<i)) printf("src");
        
next:;
        printf("\n");
    }
    close(fd);
    printf("\n");
    return 0;
}

int oss_mixer_set_from_rc(struct rcitem *cfg){
    int fd, devmask, i, vol;
    char *filename;
    char key[256];

    dbg("oss_mixer_set_from_rc()\n");
    filename = rc_str("oss_mixer_file", "/dev/mixer"); /* Path to OSS mixer */

    fd = open(filename, O_RDWR, 0);
    if (fd < 0) {
        errormsg("Failed to open %s", filename);
        return -1;
    }

    if (ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devmask) < 0){
        errormsg("Failed ioctl(SOUND_MIXER_READ_DEVMASK)");
        close(fd);
        return -1;
    }

    for (i=0; i<SOUND_MIXER_NRDEVICES; i++){
        if ((devmask & (1<<i))==0) continue;

        snprintf(key, sizeof(key), "oss_mixer_%s", cnames[i]);
        dbg("   trying key '%s'\n", key); 
        vol = rc_int(key, -1);
        if (vol == -1) continue;
        dbg("setting '%s' to %d\n", cnames[i], vol);
        
        vol = vol | (vol << 8);

        if (ioctl(fd, MIXER_WRITE(i), &vol) < 0){
            error("Failed ioctl(MIXER_WRITE(%d))", i);
            continue;
        }
    }
        
    close(fd);
    return 0;
}

void oss_dsp_info(void){
    int k;
    char filename[64];
    
    printf("\n  oss_dsp_info\n");

    oss_dsp_file_info("/dev/dsp");
        
    for (k=0; k<16; k++){
        sprintf(filename, "/dev/dsp%d", k);
        oss_dsp_file_info(filename);
    }
    
    oss_dsp_file_info("/dev/sound/dsp");
        
    for (k=0; k<16; k++){
        sprintf(filename, "/dev/sound/dsp%d", k);
        oss_dsp_file_info(filename);
    }
}

int oss_dsp_file_info(char *filename){
    int fd, caps, i, fmt, ret;
    int afmts[]={AFMT_MU_LAW, AFMT_A_LAW, AFMT_IMA_ADPCM, AFMT_U8, 
        AFMT_S16_LE, AFMT_S16_BE, AFMT_S16_NE, AFMT_S8, 
        AFMT_U16_LE, AFMT_U16_BE, AFMT_MPEG, 0};
    char *afdsc[]={"MU_LAW", "A_LAW", "IMA_ADPCM", "U8",
        "S16_LE", "S16_BE", "S16_NE", "S8",
        "U16_LE", "U16_BE", "MPEG", NULL};
    struct stat st;
    

    ret = lstat(filename, &st);
    if (ret < 0) return -1;
    if (!S_ISCHR(st.st_mode)) return -1;

    fd = open(filename, O_RDWR, 0);
    if (fd < 0) return -1;
  
    printf("%s\n", filename);
    if (ioctl(fd, SNDCTL_DSP_GETCAPS, &caps) < 0){
        errormsg("Failed ioctl(SNDCTL_DSP_GETCAPS)");
        return -1;
    }

    printf("Capsabilities: ");
    if (caps&DSP_CAP_DUPLEX) printf("DUPLEX ");
    if (caps&DSP_CAP_REALTIME) printf("REALTIME ");
    if (caps&DSP_CAP_BATCH) printf("BATCH ");
    if (caps&DSP_CAP_COPROC) printf("COPROC ");
    if (caps&DSP_CAP_TRIGGER) printf("TRIGGER ");
    if (caps&DSP_CAP_MMAP) printf("MMAP ");
    printf("\n");

    printf("Formats: ");
    if (ioctl(fd, SNDCTL_DSP_GETFMTS, &fmt) < 0){
        errormsg("Failed ioctl(SNDCTL_DSP_GETFMTS)");
        return -1;
    }
    
    for(i=0; afmts[i] != 0; i++){
        if (fmt & afmts[i]) printf("%s ", afdsc[i]);
    }
    printf("\n");
    

    


    
    close(fd);
    printf("\n");
    return 0;
}
