/*
    rc - config file functions
    Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zmem.h"
#include "error.h"
#include "list.h"
#include "rc.h"
#include "strop.h"
#include "zlinebuf.h"


struct rcitem cfg = { &cfg, &cfg };


int read_rc_file(char *filename){
    FILE *f;
    char s[1024], *c, *ss;
    int len;
    struct rcitem *rci, *rcj;

    dbg("read_rc_file('%s')\n", filename);
    f = fopen(filename, "rt");
    if (!f) return -1;

    while (fgets(s, 1000, f) != NULL){
        len = strlen(s);
        if (len > 0 && s[len - 1] == '\n'){
            s[len - 1] = '\0';
            len--;
        }
        if (len > 0 && s[len - 1] == '\r'){ 
            s[len - 1] = '\0'; /* probably unreachable, fgets() should strip \r */
            len--;
        }

		for (ss=s; *ss==' ' || *ss=='\t'; ss++); 
//	    dbg("s='%s', ss='%s'\n", s, ss);
		
        c = index(ss, '#');
		if (c) *c='\0';

        c = index(ss, ' ');
        if (!c) continue;
		if (c==ss) continue;

        rci = NULL;
        foreach(rcj, cfg){
            if (strlen(rcj->key) != c - ss) continue;
            if (strncmp(ss, rcj->key, c - ss) == 0) {
                rci = rcj;
                break;
            }
        }

        if (!rci){
            rci = z_malloc(sizeof(struct rcitem));
            rci->key = z_malloc(c - ss + 1);
            safe_strncpy(rci->key, ss, c - ss + 1);
            add_to_list(cfg, rci);
        }else{
            z_free(rci->value);
        }
        rci->value = strdup(c + 1);
	    dbg("key='%s', value='%s'\n", rci->key, rci->value);
    }
    fclose(f);    
    return 0;
}


int read_rc_files(char *name){
    char *home;
    struct zlinebuf *zlb;
    int ret;

    zlb = zlb_new(256);
    zlb_append(zlb, "/etc/");
    zlb_append(zlb, name);
    if (read_rc_file(zlb->str) == 0) return 0;

    zlb_truncate(zlb);
    home = getenv("HOME");
    if (!home) return -1;
    zlb_append(zlb, home);
    zlb_append(zlb, "/.");
    zlb_append(zlb, name);
    
    ret = read_rc_file(zlb->str);
    zlb_free(zlb);
    return 0;
}

char *rc_getopt(char *key){
    struct rcitem *rci;

    foreach(rci, cfg){
        if (strcmp(key, rci->key) != 0) continue;
//        dbg("rc_getopt('%s')=\t'%s'\n", key, rci->value);
        return rci->value;
    }
//    dbg("rc_getopt('%s') nenalezeno\n", key);
    return NULL;
}

char *rc_str(char *key, char *def){
    char *ret;
    ret = rc_getopt(key);
    if (!ret) ret = def;
    return ret;
}

int rc_int(char *key, int def){
    char *c;
    int ret;
    
    c = rc_getopt(key);
    if (!c) ret = def;
    else ret = strtol(c, NULL, 0);

    return ret;
}

/*int rc_double(char *key, double def){
    char *c;
    double ret;
    
    c = rc_getopt(key);
    if (!c) ret = def;
    else ret = atoi(c);

    return ret;
} */

char *rc_strn(char *key, int n, char *def){
    char *ret, *k;
    
    k = z_malloc(strlen(key)+20);
    sprintf(k, "%s%d", key, n);
    ret = rc_getopt(k);
//    dbg("key='%s' ret='%s'\n", k, ret);
    if (!ret) ret = def;
    z_free(k);
    return ret;
}

int rc_intn(char *key, int n, int def){
    char *c, *k;
    int ret;
    
    k = z_malloc(strlen(key)+20);
    sprintf(k, "%s%d", key, n);
    c = rc_getopt(k);
    if (!c) ret = def;
    else ret = strtol(c, NULL, 0);
    z_free(k);
    return ret;
}


void rc_info(){
    struct rcitem *rci;

    printf("\n  rc_info:\n");
    foreach(rci, cfg){
        printf("key='%s'\tvalue='%s'\n", rci->key, rci->value);
    }
}



