/*
    Real time clock/high resolution timer functions
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <string.h>
#ifdef __linux__
#include <linux/rtc.h>
#else
#include <rtc.h>
#define RTC_IRQP_SET RTCIO_IRQP_SET
#define RTC_PIE_ON RTCIO_PIE_ON
#endif

#include "error.h"
#include "rtclock.h"

static int rtc_fd = -1;

int init_rtc(const char *filename){
    const char *f = filename ? filename : "/dev/rtc";

    if ((rtc_fd = open(f, O_RDONLY)) < 0){
        errormsg("Can't open %s, using usleep", f);
        return -1;
    }
	
	unsigned long irqp = 4096;
    if (ioctl(rtc_fd, RTC_IRQP_SET, irqp) < 0) {
        errormsg("failed ioctl(RTC_IRQP_SET), using usleep");
        close(rtc_fd);
        rtc_fd = -1;
        return -1;
    }
	
    if (ioctl(rtc_fd, RTC_PIE_ON, 0) < 0) {
        errormsg("failed ioctl(RTC_PIE_ON), using usleep");
        close(rtc_fd);
        rtc_fd = -1;
        return -1;
    }
    return 0;
}


int free_rtc(){
    close(rtc_fd);
    rtc_fd = -1;
    return 0;
}

void rtc_usleep(int us){
    struct timeval start, stop, now;

    if (rtc_fd < 0){
        usleep(us);
        return;
    }

    gettimeofday(&start, NULL);
    stop.tv_usec = start.tv_usec + us;
    stop.tv_sec = start.tv_sec;
    if (stop.tv_usec > 1000000) stop.tv_usec -= 1000000, stop.tv_sec++;
            
    while(1){
        unsigned long val;
        gettimeofday(&now, NULL);
//        if (now.tv_sec < start.tv_sec) break; // time goes back
        
        if (now.tv_sec > stop.tv_sec) break;
        if (now.tv_sec == stop.tv_sec){
            if (now.tv_usec > stop.tv_usec) break;
        }
        
        if (read(rtc_fd, &val, sizeof(val)) <= 0){
            errormsg("rtc_usleep: read failed");
            break;
        }
    }
}
