
/*
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"
#include "zmem.h"

#ifndef __SOUND_H
#define __SOUND_H

#ifdef HAVE_ALSA
#ifdef HAVE_ALSA_STATIC
#include "alsa/asoundlib.h"
#else
#include <alsa/asoundlib.h>
#endif
#endif

#define ALSA_RECOVER 5

int oss_mixer_file_info(char * filename);   
int oss_dsp_file_info(char *filename);
void sound_info(void);
void alsa_info(void);

//--------------------------------------
enum sound_frame_fmt{
    SOUND_FMT_UNSET,
    SOUND_FMT_U8,
    SOUND_FMT_S16,
    SOUND_FMT_GSM,
    SOUND_FMT_ILBC
};

const char *sound_frame_fmt2str(enum sound_frame_fmt fmt);

//--------------------------------------
struct sound_fmt{
    enum sound_frame_fmt fmt;
    int channels;
    int rate;
    unsigned int per_time_us;
    unsigned int buf_time_us;

    int spf;     // sample per frame = channels
    int bps;     // bytes per sample = data width
    int bpf;     // bytes per frame = data length
    int frame_time_us;
    
    // period length
    int per_len_f;   // frames
    int per_len_s;   // samples
    int per_len_b;   // bytes

};

void sound_fmt_fill(struct sound_fmt *fmt);
snd_pcm_format_t sound_fmt2snd_pcm_format_t(struct sound_fmt *fmt);


//--------------------------------------
enum sound_mode {
    SOUND_PLAY,
    SOUND_REC,
    SOUND_DUPLEX
};

//--------------------------------------
struct sound_stream{
    struct sound_fmt *sndfmt;
    struct sound_fmt *datafmt;
    snd_pcm_t *pcm;
    snd_pcm_stream_t pcm_stream;
    int opened;
    int recover;
    char *buf;
    int sin_phase;
    // result of alsaopen
    unsigned int real_rate;
    snd_pcm_uframes_t real_buf_size, real_per_size;
    unsigned int real_buf_time_us, real_per_time_us;
};

struct sound{
    enum sound_mode mode;
    struct sound_stream *play;
    struct sound_stream *rec;
};

//--------------------------------------
struct sound *init_sound(enum sound_mode mode);
void free_sound(struct sound *sound);

struct sound_stream *init_sound_stream(void);
void free_sound_stream(struct sound_stream *stream);
void sound_format_dump(struct sound_stream *stream, struct sound_fmt *fmt); 

int sound_stream_open(struct sound_stream *stream, const char *device);
int sound_stream_close(struct sound_stream *stream);

//----------------------------

int sound_playback(struct sound_stream *stream, void *buf, int len_f);
int sound_record(struct sound_stream *stream, void *buf, int len_f);
int sound_convert(void *srcbuf, void *dstbuf, struct sound_fmt *srcfmt, struct sound_fmt *dstfmt);
int sound_avail(struct sound_stream *stream);

#endif
