/*
    usbinfo - informations about USB devices
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"

#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBFTDI_STATIC
#include "../libftdi/ftdi.h"
#else
#include <ftdi.h>
#endif

#include "usbinfo.h"


void usb_info(void){
#ifdef HAVE_LIBFTDI
    int ret;
    struct ftdi_context ftdi;
    struct ftdi_device_list *list, *dev;
    int ids[] = {0x0403, 0x6001, // FTDI USB UART
                 0x0d8c, 0x000c, // C-Media Audio Adapter
                 0xa600, 0xE110, // Davac 4.x
                 0xa600, 0xE111, // Digital analyser
                 0xa600, 0xE112, // Antenna Rotator
                 0xa600, 0xE113, // USB GPIO
                 0};
    int idi;
    char manufacturer[128], description[128], serial[128];
#ifdef DUMP_EEPROM
    unsigned char eeprom_buf[128];
#endif
    
    printf("\n  usb_info:\n");
    ret = ftdi_init(&ftdi); 
    if (ret < 0) {
        printf("error calling ftdi_init, error=%d %s", ret, ftdi.error_str);
        return;
    }
    for (idi = 0; ids[idi] != 0; idi += 2){
        ret = ftdi_usb_find_all(&ftdi, &list, ids[idi], ids[idi+1]);
        if (ret < 0) {
            printf("error calling ftdi_usb_find_all, error=%d %s", ret, ftdi.error_str);
            return;
        }
        for (dev = list; dev != NULL; dev = dev->next){
            printf("%04x:%04x  ", dev->dev->descriptor.idVendor, dev->dev->descriptor.idProduct);
            ret = ftdi_usb_get_strings(&ftdi, dev->dev, manufacturer, 128, description, 128, serial, 128);
            if (ret < 0){
                printf("Can't get strings, error=%d %s", ret, ftdi.error_str);
            }else{
                printf("%s %s ", manufacturer, description);
                if (strlen(serial)>0 && serial[0]!='?')
                    printf("%s  ", serial);
            }
            printf("\n");
#ifdef DUMP_EEPROM
            ret = ftdi_usb_open_dev(&ftdi, dev->dev);
            if (ret < 0){
                printf("Can't open device, error=%d %s", ret, ftdi.error_str);
            }else{
                ret = ftdi_read_eeprom(&ftdi, eeprom_buf);
                if (ret < 0){
                    printf("Can't open device, error=%d %s", ret, ftdi.error_str);
                }else{
                    int i;
                    unsigned short chksum_in_eeprom, chksum_computed;
                    for (i=0; i<128; i++){
                        printf("%02x ", (unsigned char)eeprom_buf[i]);
                        if (i%16==15) printf("\n");
                        
                    }
                    chksum_in_eeprom = eeprom_buf[0x7e] + 256 * eeprom_buf[0x7f];
                    chksum_computed = ftdi_checksum(eeprom_buf, 0x80);
                    printf("Checksum eeprom=%04x computed=%04x ", chksum_in_eeprom, chksum_computed);
                }
                ftdi_usb_close(&ftdi);
            }
            printf("\n");
#endif
        }
    }
    ftdi_list_free(&list);
    ftdi_deinit(&ftdi);
#else
    printf("\n  usb_info:\nlibftdi not compiled in\n\n");
#endif
}

