#include "config.h"

#include <fcntl.h>

#ifdef HAVE_LINUX_VIDEODEV_H
#include <linux/videodev.h>
#endif
#ifdef HAVE_LINUX_VIDEODEV2_H
#include <linux/videodev2.h>
#endif

#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

#include "error.h"
#include "videoinfo.h"


void video_info(void){
    int k;
    char filename[64];
    
    printf("\n  video_info\n");
    
    video_file_info("/dev/video");
    
    for (k=0; k<16; k++){
        sprintf(filename, "/dev/video%d", k);
        video_file_info(filename);
    }
    


}

int video_file_info(char *filename){
    struct stat st;
    int fd, ret, i;
    struct video_capability cap;

    ret = lstat(filename, &st);
    if (ret < 0) return -1;
    if (!S_ISCHR(st.st_mode)) return -1;

    fd = open(filename, O_RDWR);
    if (fd < 0) return -1;
    
    printf("%s\n", filename);
    if (ioctl(fd, VIDIOCGCAP, &cap) < 0){
        errormsg("ioctl(VIDIOCGCAP) failed");
        close(fd);
        return -1;
    }

    printf("name=%s\n", cap.name);
    printf("type=0x%x ", cap.type);
    if (cap.type & VID_TYPE_CAPTURE)       printf("CAPTURE ");
    if (cap.type & VID_TYPE_TUNER)         printf("TUNER ");
    if (cap.type & VID_TYPE_TELETEXT)      printf("TELETEXT ");
    if (cap.type & VID_TYPE_OVERLAY)       printf("OVERLAY ");
    if (cap.type & VID_TYPE_CHROMAKEY)     printf("CHROMAKEY ");
    if (cap.type & VID_TYPE_CLIPPING)      printf("CLIPPING ");
    if (cap.type & VID_TYPE_FRAMERAM)      printf("FRAMERAM ");
    if (cap.type & VID_TYPE_SCALES)        printf("SCALES ");
    if (cap.type & VID_TYPE_MONOCHROME)    printf("MONOCHROME ");
    if (cap.type & VID_TYPE_SUBCAPTURE)    printf("SUBCAPTURE ");
    if (cap.type & VID_TYPE_MPEG_DECODER)  printf("MPEG_DECODER ");
    if (cap.type & VID_TYPE_MPEG_ENCODER)  printf("MPEG_ENCODER ");
    if (cap.type & VID_TYPE_MJPEG_DECODER) printf("MJPEG_DECODER ");
    if (cap.type & VID_TYPE_MJPEG_ENCODER) printf("MJPEG_ENCODER ");
    printf("\n");
    printf("channels=%d\n", cap.channels);
    printf("audios=%d\n", cap.audios);
    printf("maxwidth=%d\n", cap.maxwidth);
    printf("maxheight=%d\n", cap.maxheight);
    printf("minwidth=%d\n", cap.minwidth);
    printf("minheight=%d\n", cap.minheight);

    for (i = 0; i < cap.channels; i++){
        struct video_channel ch;
        struct video_window win;
        struct video_picture pict;

        printf("channel=%d:\n", i);
        ch.channel = i;
        if (ioctl(fd, VIDIOCGCHAN, &ch) < 0){
            errormsg("\nioctl(VIDIOCGCHAN) failed");
            continue;
        }
        printf("  name=%s\n", ch.name);
        printf("  tuners=%d\n", ch.tuners);
        printf("  flags=0x%x ", ch.flags);
        if (ch.flags & VIDEO_VC_TUNER) printf("TUNER ");
        if (ch.flags & VIDEO_VC_AUDIO) printf("AUDIO ");
        printf("\n  type=0x%x ", ch.type);
        if (ch.type & VIDEO_TYPE_TV) printf("TV ");
        if (ch.type & VIDEO_TYPE_CAMERA) printf("CAMERA ");
        printf("\n  norm=0x%x\n", ch.norm);

        // select video channel
        ch.channel = i;
        if (ioctl(fd, VIDIOCSCHAN, &ch) < 0){
            errormsg("\nioctl(VIDIOCSCHAN) failed");
            continue;
        }
        if (ioctl(fd, VIDIOCGWIN, &win) < 0){
            errormsg("\nioctl(VIDIOCGWIN) failed");
            continue;
        }
        printf("  win.x=%d\n", win.x);
        printf("  win.y=%d\n", win.y);
        printf("  win.w=%d\n", win.width);
        printf("  win.h=%d\n", win.height);
        printf("  win.chromakey=0x%x\n", win.chromakey);
        printf("  win.flags=0x%x\n", win.flags);

        if (ioctl(fd, VIDIOCGPICT, &pict) < 0){
            errormsg("\nioctl(VIDIOCGPICT) failed");
            continue;
        }
        printf("  pict.brightness=%d\n", pict.brightness);
        printf("  pict.hue=%d\n", pict.hue);
        printf("  pict.colour=%d\n", pict.colour);
        printf("  pict.contrast=%d\n", pict.contrast);
        printf("  pict.whiteness=%d\n", pict.whiteness);
        printf("  pict.depth=%d\n", pict.depth);
        printf("  pict.palette=%d ", pict.palette);
        switch(pict.palette){
            case VIDEO_PALETTE_GREY: printf("GREY\n"); break;
            case VIDEO_PALETTE_HI240: printf("HI240\n"); break;
            case VIDEO_PALETTE_RGB565: printf("RGB565\n"); break;
            case VIDEO_PALETTE_RGB24: printf("RGB24\n"); break;
            case VIDEO_PALETTE_RGB32: printf("RGB32\n"); break;
            case VIDEO_PALETTE_RGB555: printf("RGB555\n"); break;
            case VIDEO_PALETTE_YUV422: printf("YUV422\n"); break;
            case VIDEO_PALETTE_YUYV: printf("YUYV\n"); break;
            case VIDEO_PALETTE_UYVY: printf("UYVY\n"); break;
            case VIDEO_PALETTE_YUV420: printf("YUV420\n"); break;
            case VIDEO_PALETTE_YUV411: printf("YUV411\n"); break;
            case VIDEO_PALETTE_RAW: printf("RAW\n"); break;
            case VIDEO_PALETTE_YUV422P: printf("YUV422P\n"); break;
            case VIDEO_PALETTE_YUV411P: printf("YUV411P\n"); break;
            case VIDEO_PALETTE_YUV420P: printf("YUV420P\n"); break;
            case VIDEO_PALETTE_YUV410P: printf("YUV410P\n"); break;
//            case VIDEO_PALETTE_PLANAR: printf("PLANAR\n"); break;
            default: printf("???\n");
        }

        printf("\n");
    }

    close(fd);

    return 0;
}
