#include "config.h"

#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>

#include "error.h"
#include "wav.h"


struct wav *wav_open(const char *filename){
    struct wav *wav;
    struct dwav_riff riff;
    struct dwav_subchunk subchunk;
    struct dwav_fmt fmt;

    memset(&fmt, 0, sizeof(fmt));
    
    wav = malloc(sizeof(struct wav));
    wav->f = fopen(filename, "rb");
    if (!wav->f){
        errormsg("Can't open %s", filename);
        free(wav);
        return NULL;
    }
    wav->for_write = 0;
                                                
    if (fread(&riff, sizeof(struct dwav_riff), 1, wav->f) != 1){
        errormsg("Can't read riff header from %s", filename);
        goto err;
    }

    if (ntohl(riff.chunk_id) != WAV_RIFF){
        error("No RIFF file %s\n", filename);
        goto err;
    }
                                            
    if (ntohl(riff.chunk_format) != WAV_WAVE){
        error("No WAVE file %s\n", filename);
        goto err;
    }

    while(1){
        if (fread(&subchunk, sizeof(struct dwav_subchunk), 1, wav->f) != 1){
            errormsg("Can't read subchunk from %s", filename);
            goto err;
        }
        
        if (ntohl(subchunk.subchunk_id) == WAV_FMT){
            if (fread(&fmt, sizeof(struct dwav_fmt), 1, wav->f) != 1){
                errormsg("Can't read fmt from %s", filename);
                goto err;
            }

            wav->bits_per_sample = ntohs(fmt.bits_per_sample);
            wav->channels = ntohs(fmt.num_channels);
            wav->rate = ntohl(fmt.sample_rate);
            
            continue;
        }
    }
    
    while(1){
        if (fread(&subchunk, sizeof(struct dwav_subchunk), 1, wav->f) != 1){
            errormsg("Can't read subchunk from %s", filename);
            goto err;
        }
        
        if (ntohl(subchunk.subchunk_id) == WAV_DATA){
            wav->bpf = wav->channels * wav->bits_per_sample / 8;
            if (!wav->bpf){
                error("Channels or bit per sample is 0, non-existent fmt chunk in %s?\n", filename);
                goto err;
            }
            
            wav->len_b = ntohl(subchunk.subchunk_size);
            wav->len_f = wav->len_b / wav->bpf;
            continue;
        }
        error("Unknown chunk 0x%08x\n", subchunk.subchunk_id);
    }
    
    return wav;
err:;
    fclose(wav->f);
    free(wav);
    return NULL;
    
}

int wav_read(struct wav *wav, void *data, int len_f){
    int ret;

    ret = fread(data, 1, len_f * wav->bpf, wav->f);
    if (ret < 0) {
        errormsg("Can't read from WAV file");
        return ret;
    }
    return ret / wav->bpf;
}


struct wav *wav_create(const char *filename){
    struct wav *wav;
//    struct dwav_riff riff;
//    struct dwav_subchunk subchunk;
//    struct dwav_fmt fmt;

    wav = malloc(sizeof(struct wav));
    wav->f = fopen(filename, "wb");
    if (!wav->f){
        errormsg("Can't open %s", filename);
        free(wav);
        return NULL;
    }
    wav->for_write = 1;

//  riff.riff = htonl(WAV_RIFF);
//  riff.
    return wav;
}

int wav_write(struct wav *wav, void *data, int len_f){
    return 0;    
}

int wav_close(struct wav *wav){
    if (wav->for_write){
#if 0        
        riff.chunk_size = htonl(
            sizeof(struct dwav_riff) - sizeof(uint32_t) /* chunk_id */ + 
            sizeof(struct dwav_subchunk) + wav.fmt_len_b +
            sizeof(struct dwav_subchubk) + wav->len_b);
#endif        
    }
    fclose(wav->f);
    free(wav);
    return 0;
}

