/*
    zlibebuf - linebuffer functions
    Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "zmem.h"
#include "error.h"
#include "zlinebuf.h"

struct zlinebuf *zlb_new(int size){
    struct zlinebuf *zlb;
    
    zlb = z_malloc(sizeof(struct zlinebuf));
    if (size <= 0) size = 1;
    zlb->size = size;
    zlb->str = z_malloc(zlb->size);
    strcpy(zlb->str, "");
    return zlb;
}

void zlb_free(struct zlinebuf *zlb){
    z_free(zlb->str);
    z_free(zlb);
}

void zlb_truncate(struct zlinebuf *zlb){
    strcpy(zlb->str, "");
}

void zlb_append(struct zlinebuf *zlb, const char *str){
    int size;
    char *c;

    size = strlen(zlb->str) + strlen(str) + 1;
    if (size > zlb->size){
        zlb->size = size;
        c = z_malloc(size);
        strcpy(c, zlb->str);
        z_free(zlb->str);
        zlb->str = c;
    }
    strcat(zlb->str, str);
}

int zlb_append_fd(struct zlinebuf *zlb, int fd, int toread){
    int size, ret, oldstrlen;
    char *c;

    oldstrlen = strlen(zlb->str);
    size = oldstrlen + toread + 1;
    if (size > zlb->size){
        zlb->size = size;
        c = z_malloc(size);
        strcpy(c, zlb->str);
        z_free(zlb->str);
        zlb->str = c;
    }
    ret = read(fd, zlb->str + oldstrlen, toread);
    if (ret < 0) return ret;
    zlb->str[oldstrlen + ret] = '\0';
    return ret;
}

struct zstring *zlb_get_line(struct zlinebuf *zlb){
    char *c;
    int i, l;
    struct zstring *zs;

    c = index(zlb->str, '\n');
    if (!c) return NULL;

    i = c - zlb->str;
    
    zs = z_malloc(sizeof(struct zstring));
    zs->str = z_malloc(i + 1);
    strncpy(zs->str, zlb->str, i);
    zs->str[i] = '\0';
    if (i > 0 && zs->str[i - 1] == '\r') zs->str[i - 1] = '\0';
    
    l = strlen(zlb->str) - i - 1;
    if (l <= 0) {
        strcpy(zlb->str, "");
    }else{
        memmove(zlb->str, zlb->str + i + 1, l + 1);
    }
//    dbg("zlb_get_line line='%s'  zlb->str='%s'\n", zs->str, zlb->str);
    return zs;
}

