/*
    zstring - string functions
    Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "zmem.h"
#include "zstring.h"
#include "error.h"

static char hexchars[] = "0123456789abcdef";


struct zstring *zs_concatesc(char *fmt, ...){
    va_list l;
    struct zstring *zs;
    
    va_start(l, fmt);
    zs = zs_vconcatesc(fmt, l);
    va_end(l);
    return zs;
}

struct zstring *zs_vconcatesc(char *fmt, va_list l){
    va_list l2;
    size_t len;
    char  *d;
    char *s, *c;
    struct zstring *zs;

#ifdef va_copy
    va_copy(l2, l);
#else
    l2 = l;
#endif
    
    s = fmt;
    len = 0;
    do {
        if (s != fmt) len++;
        for (c = s; *c != '\0'; c++){
            switch(*c){
                case ';':
                case '\\':
                case '\r':
                case '\n':
                    len += 2;
                    break;
                default:
                    len++;
                    break;
            }
        }
    }while ((s = va_arg(l, char *)) != NULL);
//    va_end(l);
    len++;

    zs = z_malloc(sizeof(struct zstring));
    zs->str = d = z_malloc(len + 2);
                                 
//    va_start(l, s1);
    s = fmt;
    len = 0;
    do {
        if (s != fmt) *d++ = ';';
        for (c = s; *c != '\0'; c++){
            switch(*c){
                case ';':
                    *d++ = '\\';
                    *d++ = ';';
                    break;
                case '\\':
                    *d++ = '\\';
                    *d++ = '\\';
                    break;
                case '\r':
                    *d++ = '\\';
                    *d++ = 'r';
                    break;
                case '\n':
                    *d++ = '\\';
                    *d++ = 'n';
                    break;
                default:
                    *d++ = *c;
                    break;
            }
        }
    }while ((s = va_arg(l2, char *)) != NULL);
#ifdef va_copy
    va_end(l2);
#endif
    *d = '\0';
    
    return zs;
}


int zs_writeconcatesc(int fd, char *fmt, ...){
    va_list l;
    struct zstring *zs;
    int ret;
    
    va_start(l, fmt);
    zs = zs_vconcatesc(fmt, l);
    va_end(l);

//    dbg("write '%s'\n", zs->str);
    ret = write(fd, zs->str, strlen(zs->str));

    zs_free(zs);
    return ret;
}



    
struct zstring *zs_vprintfesc(char *fmt, va_list l){
    va_list l2;
    size_t len;
    char  *d;
    char *c, *cc;
    struct zstring *zs;
    char sd[20];
    int ival;
    char *sval;
    int i;

#ifdef va_copy
    va_copy(l2, l);
#else
    l2 = l;
#endif

    len = 0;
//    dbg("fmt='%s'\n", fmt);
    for (c = fmt; *c != '\0'; c++){
        if (*c != '%') {
            len++;
            continue;
        }
        switch(c[1]){
            case '\0': 
                continue;
            case 'd':
                ival = va_arg(l, int);
//                dbg("d ival=%d\n", ival);
                sprintf(sd, "%d", ival);
                len += strlen(sd);
                c++;
                break; 
            case 's':
                sval = va_arg(l, char *);
//                dbg("s sval='%s'\n", sval);
                for (cc = sval; *cc != '\0'; cc++){
                    switch (*cc){
                        case ';':
                        case '\\':
                        case '\r':
                        case '\n':
                            len += 2;
                            break;
                        default:
                            len++;
                            break;
                    }
                }
                c++;
                break;
            case 'h':
                sval = va_arg(l, char *);
//                dbg("h sval='%c%c%c%c'...\n", sval[0], sval[1], sval[2], sval[3]);
                ival = va_arg(l, int);
//                dbg("h ival=%d\n", ival);
                len += 2 * ival;
                c++;
                break;
        }
    }
    
    len++; // '\0'

    zs = z_malloc(sizeof(struct zstring));
    zs->str = d = z_malloc(len);
                                 
    len = 0;
    for (c = fmt; *c != '\0'; c++){
        if (*c != '%') {
            *d++ = *c;
            continue;
        }
        switch(c[1]){
            case '\0': 
                continue;
            case 'd':
                ival = va_arg(l2, int);
                sprintf(d, "%d", ival);
                d += strlen(d);
                c++;
                break; 
            case 's':
                sval = va_arg(l2, char *);
                for (cc = sval; *cc != '\0'; cc++){
                    switch (*cc){
                        case ';':
                        case '\\':
                        case '\r':
                        case '\n':
                            *d++ = '\\';
                            break;
                    }
                    *d++ = *cc;
                }
                c++;
                break;
            case 'h':
                sval = va_arg(l2, char *);
                ival = va_arg(l2, int);
                for (i = ival, cc = sval; 
                        i > 0; 
                        i--, cc++){

                    *d++ = hexchars[(*cc >> 4) & 0x0f];
                    *d++ = hexchars[*cc & 0x0f];
                }
                c++;
                break;
        }
    }
#ifdef va_copy
    va_end(l2);
#endif
    *d = '\0';
    
    return zs;
}

struct zstring *zs_printfesc(char *fmt, ...){
    va_list l;
    struct zstring *zs;
    
    va_start(l, fmt);
    zs = zs_vprintfesc(fmt, l);
    va_end(l);
    return zs;
}


int zs_writeprintfesc(int fd, char *fmt, ...){
    va_list l;
    struct zstring *zs;
    int ret;
    
    va_start(l, fmt);
    zs = zs_vprintfesc(fmt, l);
    va_end(l);

    ret = write(fd, zs->str, strlen(zs->str));

    zs_free(zs);
    return ret;
}





struct zstring *zs_strdup(const char *str){
    int len;
    struct zstring *zs;

    len = strlen(str);
    zs = z_malloc(sizeof(struct zstring));
    zs->str = z_malloc(len + 1);
    memcpy(zs->str, str, len + 1);
    return zs;
}

char *zs_tokenize(struct zstring *zstr, int first){
    int backslash=0;
    int i;
    char c, *d, *buf;
    int len;

    if (first) zstr->tokenpos = 0;
    len = strlen(zstr->str + zstr->tokenpos) + zstr->tokenpos;
    if (zstr->tokenpos > len) return NULL;
    if (zstr->tokenpos < 0) return NULL;
    buf = d = zstr->str + zstr->tokenpos;

    for (i = zstr->tokenpos; i < len; i++){
        c = zstr->str[i];
        if (backslash){
            backslash = 0;
            switch(c){
                case '\\':
                case ';':
                    break;
                case 'r':
                    c = '\r';
                    break;
                case 'n':
                    c = '\n';
                    break;
                default: /* error */
                    fprintf(stderr, "bad escape \\%c\n", c);
                    break;
            }
        }else{
            switch (c){
                case '\\':
                    backslash = 1;
                    continue;
                case ';':
                    zstr->tokenpos = i + 1;
                    *d='\0';
                    return buf;
                default:
                    break;
            }
        }
        *d++ = c;
    }
    zstr->tokenpos = -1;
    *d='\0';
    return buf;
}

void zs_free(struct zstring *zstr){
    z_free(zstr->str);
    zstr->str = NULL;
    z_free(zstr);
}

int zs_tokens(struct zstring *zstr){
    int backslash=0;
    int tokens=1;
    char *c;

    for (c = zstr->str; *c != '\0'; c++){
        if (backslash){
            backslash = 0;
            continue;
        }

        switch (*c){
            case '\\':
                backslash = 1;
                break;
            case ';':
                tokens++;
                break;
        }
    }
    return tokens;
}

/*int zwrite(int fd, struct zstring *zs){
	int ret;
    char *c;
    
    c = g_strconcat(zs->str, "\n", NULL);
	zfree(zs);
	ret = write(fd, c, strlen(c));
    z_free(c);
	return ret;
} */




struct zstring *zs_import_bin(const char *bin, int len){
    struct zstring *zs;
    int i;
    const char *c;
    char *d;
    
    zs = z_malloc(sizeof(struct zstring));
    zs->size = len * 2 + 1;   
    zs->str = z_malloc(zs->size);
    
    for (i = len, c = bin, d = zs->str; 
         i > 0 ;
         i--, c++){

        *d++ = hexchars[(*c >> 4) & 0x0f];
        *d++ = hexchars[*c & 0x0f];
    }
    *d = '\0';
    return zs;
}

static char hexval(const char c){
    if (c >= '0' && c <= '9') return c - '0';
    if (c >= 'a' && c <= 'f') return c - 'a' + 10;
    if (c >= 'A' && c <= 'F') return c - 'a' + 10;
    return 0;
}

// returned binary data is in zs->str and zs->size
struct zstring *zs_import_hex(const char *str){
    const char *c;
    char *d;
    struct zstring *zs;

    zs = z_malloc(sizeof(struct zstring));
    zs->size = strlen(str) / 2 + 2;   
    zs->str = z_malloc(zs->size);
    zs->size = 0;

    for (c = str, d = zs->str; 
         *c != '\0'; 
         c++){

        *d = hexval(*c++);
        *d <<= 4;
        *d |= hexval(*c);
        d++;
        zs->size++;
    }
    return zs;
}

#if 0
int main(){
    char *c;
    zchar *z;
    int i;

    c = zconcatesc("aa", ";", "", NULL);
    printf("c='%s'\n", c);
    z = zstrdup(c);
    printf("z='%s'\n", z);

    i = 0;
    for (c = ztokenize(z, 1); c != NULL; c = ztokenize(z, 0)){
        printf("%d='%s'\n", i, c);
        i++;
    }
    zfree(z);
    return 0;
}
#endif
