/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

int init_cftdi(struct crig *crig){
    int ftdi_vid, ftdi_pid;
    int ret, f, ftdi_baudrate, i;
    char *ftdi_serial;
    int verbose = 1;
    char *ftdi_prot;
//    char errbuf[1024];
//  char *device;
    
    dbg("init_cftdi\n");

    crig->crig_free   = free_cftdi;
    crig->crig_reset  = cftdi_reset;
    crig->crig_gpo    = cftdi_gpo;
    crig->crig_set_output = cftdi_set_output;
    crig->crig_force_send_gpi    = cftdi_force_send_gpi;

    if (pipe(crig->ftdi_pipe)) internal("Can't create cftdi pipe");
    crig->ftdi_zlb = zlb_new(256);
    
    for (i=0; i < MAXGPO; i++) crig->gpo[i] = 1;
    
    for (f = 0; f < MAXFTDI; f++){
        struct ftdi_context *ftdi;

        ftdi_vid = rc_intn("cftdi_vid", f, 0); /* USB vendor ID */
        ftdi_pid = rc_intn("cftdi_pid",  f,0); /* USB product ID */
        if (!ftdi_vid || !ftdi_pid) continue;

        ftdi_serial = rc_strn("cftdi_serial", f, ""); /* FTDI serial number */
        ftdi_baudrate = rc_intn("cftdi_baudrate", f, 300); /* FTDI baud rate (300-3000000) */
        crig->ftdi_bitmode[f] = rc_intn("cftdi_bitmode", f, 1); /* FTDI bitmode (0=serial, 1=bitbang etc) */
        ftdi_prot = rc_strn("cftdi_prot", f, ""); /* FTDI serial protocol (ic706p) */
        if (ftdi_prot && *ftdi_prot && strcmp(ftdi_prot, "ic706p")==0) crig->ftdi_prot[f] = PROT_IC706P;
        dbg("ftdi_prot='%s'\n", ftdi_prot);
        crig->ftdi_bitmask[f] = 0x00;
        crig->ftdi_gpo[f] = 0xff;
       // ftdi_dummy = rc_int("ftdi_dummy", 0); /* Dummy mode without FTDI device */

        if (!ftdi_vid) {
            error("Refuse to open VID=%04x, probably incorrect configuration.\n", ftdi_vid);
            return 1;
        }

        dbg("ftdi_open %04x:%04x serial='%s' server=%d\n", ftdi_vid, ftdi_pid, ftdi_serial, opt_s);
                                    
//        if (ftdi != NULL) return 0;
        pthread_mutex_init(&crig->ftdi_gpi_mutex[f], NULL);

        crig->ftdi[f] = ftdi = ftdi_new();
        if (!ftdi){
            if (verbose) error("Can't init ftdi library for ftdi");
            return 1;
        }
        if (ftdi_serial && *ftdi_serial)
            ret=ftdi_usb_open_desc(ftdi, ftdi_vid, ftdi_pid, NULL, ftdi_serial);
        else
            ret=ftdi_usb_open(ftdi, ftdi_vid, ftdi_pid);
        
        if (ret){
            if (verbose) {
                error("Can't open ftdi device %04x:%04x, error=%d %s\n", ftdi_vid, ftdi_pid, ret, ftdi->error_str);
                if (ret==-8) error("Maybe run program as root\n");
            }
            free_cftdi(crig);
            return ret;
        }
        
        ret=ftdi_set_baudrate(ftdi, ftdi_baudrate);
        if (ret){
            if (verbose) error("Can't set baudrate for ftdi, error=%d %s\n", ret, ftdi->error_str);
            free_cftdi(crig);
            return ret;
        }
        
    }
        
/*    ret=ftdi_write_data_set_chunksize(ftdi, 50);
    if (ret){
        if (verbose) error("Can't set chunk size, error=%d %s\n", ret, ftdi->error_str);
        free_ftdi();
        return ret;
    }*/

//    crig->sdev = init_sdev();
//    device = rc_getopt("ftdi_device"); /* Path to ftdi device (/dev/ttyS0) */
/*    if (!device) device="/dev/ttyS0";
    rig->sdev->device=strdup(device);
    rig->sdev->callback=ftdi_async;
    rig->sdev->callarg=rig;
    sdev_open(rig->sdev);*/
    
    crig_reset();

    set_handlers(crig->ftdi_pipe[0], cftdi_read_handler, NULL, NULL, crig);
    pthread_create(&crig->ftdi_async_thread, NULL, cftdi_async_thread_func, crig);

    return 0;
}

void free_cftdi(struct crig *crig){
//    char errbuf[1024];
    int f;
    
    if (!crig) return;
    
    if (crig->ftdi_async_thread) pthread_cancel(crig->ftdi_async_thread);
    close(crig->ftdi_pipe[0]);
    close(crig->ftdi_pipe[1]);
    zlb_free(crig->ftdi_zlb);
    for (f = 0; f < MAXFTDI; f++){
        struct ftdi_context *ftdi;

        ftdi = crig->ftdi[f];
        if (!ftdi) continue;
        pthread_mutex_destroy(&crig->ftdi_gpi_mutex[f]);
        ftdi_free(ftdi);
        crig->ftdi[f] = NULL;
    }    
	
	return;
}

int cftdi_reset(struct crig *crig){

    return 0; 
}


int cftdi_set_output(struct crig *crig, int gponr, int output){
    int f = gponr >> 3;
    int b = gponr & 0x07;
    struct ftdi_context *ftdi = crig->ftdi[f];
    int ret;
    
    if (gponr == IO_FLUSH){
        for (f = 0; f < MAXFTDI; f++){
            ftdi = crig->ftdi[f];
            if (!ftdi) continue;

            ret=ftdi_set_bitmode(ftdi, crig->ftdi_bitmask[f], crig->ftdi_bitmode[f]);
            dbg("ftdi_set_bitmode(%d, 0x%02x, %d)=%d\n", f, crig->ftdi_bitmask[f], crig->ftdi_bitmode[f], ret);
            if (ret){
                error("Can't enable bitbang, error=%d %s\n", ret, ftdi->error_str);
                return ret;
            }
            
            crig->ftdi_gpo[f] |= ~crig->ftdi_bitmask[f];
            ret = ftdi_write_data(ftdi, (unsigned char *)&crig->ftdi_gpo[f], 1);
            if (ret!=1) {
                error("Can't write, error=%d %s\n", ret, ftdi->error_str);
            }
        }
        return 0;
    }
    
    if (!ftdi) return -1;

    if (output) crig->ftdi_bitmask[f] |= 1 << b;
    else        crig->ftdi_bitmask[f] &= ~(1 << b);

    return 0;
}


int cftdi_gpo(struct crig *crig, int gponr, int on){
    int ret;
    int f = gponr >> 3;
    int b = gponr & 0x07;
    struct ftdi_context *ftdi = crig->ftdi[f];
    
    if (!ftdi) return -1;

    if (on) {
        crig->ftdi_gpo[f] |= 1 << b;
        crig->gpo[gponr] = 1;
    }else{
        crig->ftdi_gpo[f] &= ~(1 << b);
        crig->gpo[gponr] = 0;
    }

//    dbg("cftdi_gpo: write(%02x)\n", (unsigned char)crig->ftdi_gpo[f]);
    ret = ftdi_write_data(ftdi, (unsigned char *)&crig->ftdi_gpo[f], 1);
    if (ret!=1) {
        error("Can't write, error=%d %s\n", ret, ftdi->error_str);
    }
    
    return 0;
}

int cftdi_force_send_gpi(struct crig *crig){
    int f;
    struct ftdi_context *ftdi;
    
    for (f = 0; f < MAXFTDI; f++){
        ftdi = crig->ftdi[f];
        if (!ftdi) continue;
    
        pthread_mutex_lock(&crig->ftdi_gpi_mutex[f]);
        crig->ftdi_gpi[f] ^= 0xff;
        pthread_mutex_unlock(&crig->ftdi_gpi_mutex[f]);
    }
    return 0;
}


void *cftdi_async_thread_func(void *arg){
    struct crig *crig = (struct crig *)arg;
    int n, ret, i, f, b;
    struct ftdi_context *ftdi;
    char data[256];// same as data2[256] in cprot_ic706p_read
    char s[50];

    while(1){
//        dbg("cftdi_async_thread_func\n");
        pthread_testcancel();
        n=0;
        for (f = 0; f < MAXFTDI; f++){
            ftdi = crig->ftdi[f];
            if (!ftdi) break;
            ret = ftdi_read_data(ftdi, (unsigned char *)data, sizeof(data));
           // dbg("ret%d=%d\n",f, ret);
            if (ret<=0) continue;
            n += ret;

            switch(crig->ftdi_bitmode[f]){
                case BITMODE_RESET:
                    if (crig->ftdi_prot[f] == PROT_IC706P){
                        cprot_ic706p_read(crig, f, data, ret);
                    }else{
                        dbg("rda%d(%d): ", f, ret);
                        for (i=0; i<ret; i++) {dbg("%02x ", (unsigned char)data[i]); if (i>40) break;}
                        dbg("\n");
                        zs_writeprintfesc(crig->ftdi_pipe[1], "%d;%h\n", f, data, ret);
                    }
                    break;
                case BITMODE_BITBANG:
                    for (i=0; i<ret; i++){
                        char xor;
                        pthread_mutex_lock(&crig->ftdi_gpi_mutex[f]);
                        xor = data[i] ^ crig->ftdi_gpi[f];
                        xor &= ~crig->ftdi_bitmask[f];
                        if (xor){
//                        dbg("xor=%02x bitmask=%02x\n", (unsigned char)xor, (unsigned char)crig->ftdi_bitmask[f]);
                            crig->ftdi_gpi[f] = data[i];
                            for (b = 0; b < 8; b++){
                                if (!(xor & (1<<b))) continue;
                                sprintf(s, "%d;%d;%d\n", f, b, !!(data[i] & (1 << b)));
                                write(crig->ftdi_pipe[1], s, strlen(s));
                            }
                        }
                        pthread_mutex_unlock(&crig->ftdi_gpi_mutex[f]);
                    }
                    break;
            }
        }
        if (!n) usleep(10000);
    }

    return NULL;
}

void cftdi_read_handler(void *xxx){
    int ret;
    char buf[FBUFLEN+1];
    struct zstring *zs, *zs2;
    struct crig *crig = (struct crig *)xxx;
    char *c_f;
    
    ret=read(crig->ftdi_pipe[0], buf, FBUFLEN);
//    dbg("\ncftdi_read_handler ret=%d\n", ret);
    if (ret <= 0) internal("Can't read from cftdi pipe");
    buf[ret] = '\0';
    zlb_append(crig->ftdi_zlb, buf);
//    dbg("zlb='%s'\n", crig->ftdi_zlb->str);

    while ((zs=zlb_get_line(crig->ftdi_zlb)) != NULL){
        int f, b, v;
        char *c;

//        dbg("zs='%s'\n", zs->str);
        c_f = zs_tokenize(zs, 1);
        f = atoi(c_f);
        switch(crig->ftdi_bitmode[f]){
            case BITMODE_RESET:
                c = zs_tokenize(zs, 0);
                zs2 = zs_printfesc("SER;%d;%s", f, c);
                rel_write(sconn, zs2->str);
                zs_free(zs2);
                break;    
            case BITMODE_BITBANG:
                b = atoi(zs_tokenize(zs, 0));
                v = atoi(zs_tokenize(zs, 0));
        //        dbg("cftdi_read_handler(f=%d, b=%d, v=%d)\n", f, b, v);
                zs_free(zs);
                crig_async_gpi(crig, f * 8 + b, v); 
                break;
        }
    }
}

int cftdi_write(struct crig *crig, int f, char *data, int len){
    int ret;

    /*int i;
    dbg("wr%d(%d): ", f, len);
    for (i=0; i<data; i++) {dbg("%02x ", (unsigned char)data[i]); if (i>40) break;}
    dbg("\n");*/

    ret = ftdi_write_data(crig->ftdi[f], (unsigned char *)data, len);
    return ret;
}

// very simle code - no problem when some keepalive is not catched
void cprot_ic706p_read(struct crig *crig, int f, char *data, int len){
    int i, l;
    char data2[256];

#if 0    
    dbg("rdb%d(%d): ", f, len);
    for (i=0; i<len; i++) {dbg("%02x ", (unsigned char)data[i]); if (i>40) break;}
    dbg("\n");
#endif

    l = 0;
    for (i=0; i < len; i++){

        if (i < len - 3 &&
            data[i+0]== (char)0xfe &&
            data[i+1]== (char)0x0b &&
            data[i+2]== (char)0x00 &&
            data[i+3]== (char)0xfd){
            // keepalive packet
            //dbg("drop keepalive\n");
            i += 3;
            continue;
        }
        data2[l++] = data[i];
    }   

/*    memcpy(data2, data, 256);
    l = len;*/
    if (l <= 0) return; 

#if 1
    dbg("rdc%d(%d): ", f, l);
    for (i=0; i<l; i++) {dbg("%02x ", (unsigned char)data2[i]); if (i>40) break;}
    dbg("\n");
#endif
    zs_writeprintfesc(crig->ftdi_pipe[1], "%d;%h\n", f, data2, l);
}


