/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct crig *crig;

char lcd[20];

int init_crig(){
    char *crigtype;
    int ret;

    crigtype = rc_getopt("crig_type"); /* Type of crig (ppdev, sior, siop) */
    if (!crigtype){
        error("crig_type unspecified\n");
        return -1;
    }
	crig = mem_alloc(sizeof(struct crig));
	memset(crig, 0, sizeof(struct crig));
    
/*TODO    if (strcasecmp(crigtype, "siop")==0) {
        ret = init_siop(crig);
        if (ret) return ret;
    } */
    if (strcasecmp(crigtype, "cftdi")==0) {
        ret = init_cftdi(crig);
        if (ret) return ret;
    }

    if (crig->crig_set_output){
        crig->crig_set_output(crig, IO_PWR, 0);
        crig->crig_set_output(crig, IO_PWK, 0);
        crig->crig_set_output(crig, IO_8V,  1);
        crig->crig_set_output(crig, IO_CW,  0);
        crig->crig_set_output(crig, IO_PTT, 0);
        crig->crig_set_output(crig, IO_PA,  0);
        crig->crig_set_output(crig, IO_LNA, 0);
        crig->crig_set_output(crig, 7, 0);
        
        crig->crig_set_output(crig, IO_FLUSH, 0);
    }
    strcpy(lcd, "");
    return 0;
}

void free_crig(){
	mem_free(crig);
}

int crig_reset(void){
    if (!crig) return -1;

    if (!crig->crig_reset) return 0;

    
    return crig->crig_reset(crig);
}

void crig_grab(){ 
    dbg("crig_grab %d\n", crig->grab);
}

int crig_gpo(int gponr, int on){
    if (!crig) return -1;

    if (!crig->crig_gpo) return 0;
    return crig->crig_gpo(crig, gponr, on);
}

int crig_force_send_gpi(){
    if (!crig) return -1;

    dbg("crig_force_send_gpi()\n");
    if (!crig->crig_force_send_gpi) return 0;
    return crig->crig_force_send_gpi(crig);
}

void crig_async_gpi(struct crig *crig, int gpinr, int on){ /* called from callback by "device driver" */
    char s[256];
    dbg("crig_async_gpi(nr=%d, %d)\n", gpinr, on);
   
    switch (gpinr){
        case IO_PWR:
        case IO_PWK:
        case IO_CW:
        case IO_PTT:
        case IO_PA:
        case IO_LNA:
            sprintf(s, "GPO;%d;%d",gpinr, on); 
            rel_write(sconn, s);
            break;
    }
    return;
}

void crig_async_elbug(struct crig *crig, int gpinr, int on){ /* called from callback by "device driver" */
    dbg("crig_async_elbug('%c', %d)\n", gpinr==15?'-':'.', on);

}

void crig_async_key(struct crig *crig, unsigned char key){ /* called from callback by "device driver" */
    char s[30];
    int ret;

    dbg("crig_async_key('%c', %s)\n", key&0x7f, key&0x80?"release":"press");

    if (key&0x80) return;
    if (key=='C'){
        if (strlen(lcd)>0){
            lcd[strlen(lcd)-1]='\0';
        }
    }else{
        if (strlen(lcd)==16){
            memmove(lcd, lcd+1, 16);
        }
        lcd[strlen(lcd)+1]='\0';
        lcd[strlen(lcd)]=key;
    }
    sprintf(s, "%-16s", lcd);
	if (crig->crig_lcd){
		ret=crig->crig_lcd(crig, 0x40, s);   
		if (ret<0){
			dbg("Can't set LCD text error=%d\n\n", ret); 
			return;
		}
	}
}
