/*
    iprig - ham radio via internet
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>
    and authors of web browser Links 0.96

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#define LEAK_DEBUG_LISTx
#include "../config.h"

// for strndup
#define _GNU_SOURCE

#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#include <stdarg.h>

#ifdef HAVE_LINUX_PPDEV_H
#include <linux/ppdev.h>
#include <linux/parport.h>
#endif


#ifdef HAVE_LIBFTDI_STATIC
#include "../libftdi/ftdi.h"
#else
#ifdef HAVE_LIBFTDI
#include <ftdi.h>
#endif
#endif


//#include <sys/soundcard.h>

//#include "../libzia/codec.h"
#include "../libzia/error.h"
#include "../libzia/list.h"
#include "../libzia/md5.h"
#include "../libzia/misc.h"
#include "../libzia/rc.h"
#include "../libzia/sound.h"
#include "../libzia/usbinfo.h"
#include "../libzia/videoinfo.h"
#include "../libzia/zlinebuf.h"
#include "../libzia/zmutex.h"
#include "../libzia/zstring.h"

#define mem_alloc malloc
#define mem_free free
#define mem_realloc realloc

#ifdef HAVE_ALSA
#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include <alsa/asoundlib.h>
#else
typedef long snd_pcm_sframes_t;
#endif

#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif

struct srig; 

#if SIZEOF_LONG == SIZEOF_VOID_P
typedef long int vint;
#else
typedef int vint;
#endif

/* main.c */
extern int opt_c, opt_d, opt_e, opt_h, opt_i, opt_m, opt_p, opt_s, opt_t, opt_u, opt_v;
extern char *opt_hostname;
extern char *opt_test;

void init(void);
void terminate_all_subsystems(void);
char *parse_options(int argc, char **argv);
void usage(void);
void arg0(char *a0);


/* select.c */
#define H_READ  0
#define H_WRITE 1
#define H_ERROR 2
#define H_DATA  3

typedef long ttime;


extern int terminate;

void select_loop(void (*init)(void));
#ifdef LEAK_DEBUG_LIST
void debug_set_handlers(char *file, int line, int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
int debug_install_timer(char *file, int line, ttime t, void (*func)(void *), void * cba);
void debug_kill_timer(char *file, int line, int id);
#else
void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void * cba);
int install_timer(ttime t, void (*func)(void *), void * cba);
void kill_timer(int id);
#endif

#ifdef LEAK_DEBUG_LIST
#define set_handlers(n, r, w, x, a) debug_set_handlers(__FILE__, __LINE__, n, r, w, x, a)
#define install_timer(x, y, z) debug_install_timer(__FILE__, __LINE__, x, y, z)
#define kill_timer(x) debug_kill_timer(__FILE__, __LINE__, x)
#endif

void check_bottom_halves(void);

void normalize_timeval(struct timeval *tv);
void install_signal_handler(int sig, void (*fn)(void *), void * cba, int critical);
    
extern struct timeval start;
#define ST_START gettimeofday(&start, NULL)
#define ST_STOP {\
    struct timeval stop;\
    int sec, usec;\
    gettimeofday(&stop, NULL);\
    usec=stop.tv_usec-start.tv_usec;\
    sec=stop.tv_sec-start.tv_sec;\
    if (usec<0){usec+=1000000;sec--;}\
    dbg("stopky: %s: %d.%06d \n", __FUNCTION__, sec,usec);\
}

/* tcp.c */

#define MAXRDBUF 1024
#define MAXWRBUF 1024
/* milliseconds */
#define CONNECTION_TIMEOUT 30000
#define SEND_ACK_TIMEOUT 10000
#define ACK_TIMEOUT 20000
#define REMOVE_TIMEOUT 10000

extern int tcp_port, rudp_port, sudp_port;

enum tcp_state { 
    TCP_INIT=0, 
    TCP_CONNECTING,
    TCP_CONNECTED,
    TCP_WAIT_ACK,
    TCP_DISCONNECTED,
    TCP_DEAD,  /* conn removed immediately */
};

struct conn {
    struct conn *next;  
    struct conn *prev;
    int sock; 
    enum tcp_state state;
    struct sockaddr_in sin, udpsin;
    int relseq;
    int timer;
//    char rdbuf[MAXRDBUF];
//    int rdbuflen;
    struct zlinebuf *rdbuf2;    
    char wrbuf[MAXWRBUF];
    int wrbuflen;
    int disconnecting;
    char *salt, *user;
    int authorized;
    unsigned long int ssrc;
};

struct conn conn_list;  	   		/* list of server's connection, valid on server */
ZMUTEX_DEFINE(conn_list);
extern struct conn *gconn;      /* connection to client with grabbed control */
extern struct conn *sconn; 		/* connection to server, valid only on client */
ZMUTEX_DEFINE(sconn);


int init_tcp(void);
void tcp_accept_handler(void *xxx);
void tcp_read_handler(void *xxx);
void tcp_exception_handler(void *xxx);
int tcp_send_state(struct conn *conn);
void tcp_disconnect(struct conn *conn, char *reason);
int conn_prod_state(struct conn *conn);
void tcp_set_state(struct conn *conn, enum tcp_state state);
char *tcp_state_s(enum tcp_state state);
void kill_conn_timer(void *xxx);
void send_ack_timer(void *xxx);
void remove_conn_timer(void *xxx);
void free_conn(struct conn *conn);
void tcp_connect(char *hostname, int port);
void tcp_connected_handler(void *xxx);
void after_connect(struct conn *conn);
void rel_write(struct conn *conn, char *s);
void rel_write_all(char *s);
void rel_read(struct conn *conn, char *line);
void rel_sread(struct conn *conn, char *cmd, struct zstring *zs);
void rel_cread(struct conn *conn, char *cmd, struct zstring *zs);
void send_grab(struct conn *conn, int grab);
void client_authorized(struct conn *conn);


/* sdev.c */

/* C5 FCE ADR LEN DATA CHK */

struct sdev{
	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct conn *conn;
    int timeout_ms;
    char *device;
    void (*callback)(void *owner, char fce, char saddr, char *data, unsigned char len);
    void *callarg;
};


struct sdev *init_sdev(void);
void free_sdev(struct sdev *sdev);
int sdev_open(struct sdev *sdev);
int sdev_close(struct sdev *sdev);
char sd_chk(char *s, int len);
int sdev_send(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
int sdev_recv(struct sdev *sd, char *data, unsigned char len);
int sdev_prot(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
void sdev_read_handler(void *xxx);
void sdev_async(struct sdev *sd, char fce, char saddr, char *data, unsigned char len);
const char *sdev_error(int err);


/* term.c */

#define KBD_ENTER   0x100
#define KBD_BS      0x101
#define KBD_TAB     0x102
#define KBD_ESC     0x103
#define KBD_LEFT    0x104
#define KBD_RIGHT   0x105
#define KBD_UP      0x106
#define KBD_DOWN    0x107
#define KBD_INS     0x108
#define KBD_DEL     0x109
#define KBD_HOME    0x10a
#define KBD_END     0x10b
#define KBD_PGUP    0x10c
#define KBD_PGDN    0x10d

#define KBD_F1      0x120
#define KBD_F2      0x121
#define KBD_F3      0x122
#define KBD_F4      0x123
#define KBD_F5      0x124
#define KBD_F6      0x125
#define KBD_F7      0x126
#define KBD_F8      0x127
#define KBD_F9      0x128
#define KBD_F10     0x129
#define KBD_F11     0x12a
#define KBD_F12     0x12b

#define KBD_CTRL_C  0x200

#define KBD_SHIFT   1
#define KBD_CTRL    2
#define KBD_ALT     4

#define ESC_TIMEOUT         200

struct term{
  	int fd;
	char rdbuf[MAXRDBUF];
	int rdbuflen;
	struct termios oldtermios;
	int timer;
};


extern struct term *term;

int init_term(void);
void free_term(void);
void term_read_handler(void *xxx);
void term_exception_handler(void *xxx);
int get_esc_code(char *str, int len, char *code, int *num, int *el);
int process_queue(void);
void kbd_timeout(void *);
int handle_key(int key, int shifts);

/* crig.c */

#define MAXGPO 32
#define MAXGPI 32
#define MAXFTDI 4

#define IO_FLUSH -1

#define IO_PWR  0   // client -> server
#define IO_PTT  1   // client -> server
#define IO_8V   2   // server -> client
#define IO_IO0  3   // client -> server
#define IO_PWK  4   // client -> server
#define IO_PA   5   // client -> server
#define IO_LNA  6   // client -> server
#define IO_CW   7   // client -> server

#define RIO_SDA 8
#define RIO_SCL 9

#define PROT_IC706P 0x1

extern struct crig *crig;

struct crig{
    int grab;
//	int ptt, cws, pwr, lna, pa, pwk;
    int gpo[MAXGPO];
    int gpi[MAXGPI];
    
    void (*crig_free)(struct crig *crig);
    int (*crig_reset)(struct crig *crig);

    /* digital outputs */
//    int (*crig_8v)(struct crig *crig, int on);
    int (*crig_set_output)(struct crig *crig, int gponr, int out);
    int (*crig_gpo)(struct crig *crig, int gponr, int on);
    int (*crig_force_send_gpi)(struct crig *crig);
    int (*crig_lcd)(struct crig *crig, unsigned char addr, char *text);
    
    /* digital inputs */
    int (*crig_gpi)(struct crig *crig, int gpinr);
//  int (*crig_key)(struct crig *crig, unsigned char key);
    
	/* siop */
	struct sdev *sdev;
	
	/* ftdi */
	struct ftdi_context *ftdi[MAXFTDI];
    char ftdi_gpo[MAXFTDI];
    char ftdi_gpi[MAXFTDI];
    pthread_mutex_t ftdi_gpo_mutex[MAXFTDI];
    pthread_mutex_t ftdi_gpi_mutex[MAXFTDI];
    pthread_t ftdi_async_thread;
    int ftdi_bitmode[MAXFTDI];
    int ftdi_bitmask[MAXFTDI];
    int ftdi_pipe[2];
    struct zlinebuf *ftdi_zlb;
    int ftdi_prot[MAXFTDI];
};

int init_crig(void);
void free_crig(void);
int crig_reset(void);
void crig_grab(void);
void crig_async_gpi(struct crig *crig, int gpinr, int on);
int crig_gpo(int gponr, int on);
int crig_force_send_gpi(void);

void cprot_ic706p_read(struct crig *crig, int f, char *data, int len);
int cftdi_write(struct crig *crig, int f, char *data, int len);

/* srig.c */

extern struct srig *srig;

struct srig{
    int grab;
//	int ptt, cws, pwr, lna, pa, pwk;
    int gpo[MAXGPO];
    int gpi[MAXGPI];
    
    void (*srig_free)(struct srig *srig);
    int (*srig_reset)(struct srig *srig);

    /* digital outputs */
//    int (*srig_8v)(struct srig *srig, int on);
    int (*srig_set_output)(struct srig *srig, int gponr, int output);
    int (*srig_gpo)(struct srig *srig, int gponr, int on);
    int (*srig_force_send_gpi)(struct srig *srig);
    int (*srig_lcd)(struct srig *srig, unsigned char addr, char *text);
    
    /* digital inputs */
    int (*srig_gpi)(struct srig *srig, int gpinr);
//  int (*srig_key)(struct srig *srig, unsigned char key);

	/* ppdev */
	int ppdev_fd;
	
	/* siop */
	struct sdev *sdev;
	
	/* ftdi */
	struct ftdi_context *ftdi[MAXFTDI];
    char ftdi_gpo[MAXFTDI];
    char ftdi_gpi[MAXFTDI];
    pthread_mutex_t ftdi_gpo_mutex[MAXFTDI];
    pthread_mutex_t ftdi_gpi_mutex[MAXFTDI];
    pthread_t ftdi_async_thread;
    pthread_t ftdi_ic706p_thread;
    pthread_mutex_t ftdi_ic706p_mutex[MAXFTDI];
    int ftdi_bitmode[MAXFTDI];
    int ftdi_bitmask[MAXFTDI];
    int ftdi_pipe[2];
    struct zlinebuf *ftdi_zlb;
    char ftdi_ic706p_last[MAXFTDI];
    int ftdi_prot[MAXFTDI];
};

int init_srig(void);
void free_srig(void);
int srig_reset(void);
int srig_gpo(int gponr, int on);
void srig_async_gpi(struct srig *srig, int gpinr, int on);

void srig_grab(void);
int srig_force_send_gpi(void);

int sftdi_write(struct srig *srig, int f, char *data, int len);
void *sftdi_ic706p_thread_func(void *arg);

/* ppdev.c */

int init_ppdev(struct srig *srig);
void free_ppdev(struct srig *srig);
int ppdev_reset(struct srig *srig);

	/* digital outputs */
int ppdev_ptt(struct srig *srig, int on);
int ppdev_cw(struct srig *srig, int on);



/* cftdi.c */

#define FBUFLEN 256

int init_cftdi(struct crig *crig);
void free_cftdi(struct crig *crig);
int cftdi_reset(struct crig *crig);

int cftdi_set_output(struct crig *crig, int gponr, int output);
int cftdi_gpo(struct crig *crig, int gponr, int on);
void *cftdi_async_thread_func(void *);
void cftdi_read_handler(void *xxx);
int cftdi_force_send_gpi(struct crig *crig);

/* sftdi.c */
            
int init_sftdi(struct srig *srig);
void free_sftdi(struct srig *srig);
int sftdi_reset(struct srig *srig);

int sftdi_set_output(struct srig *srig, int gponr, int output);
int sftdi_gpo(struct srig *srig, int gponr, int on);
void *sftdi_async_thread_func(void *);
void sftdi_read_handler(void *xxx);
int sftdi_force_send_gpi(struct srig *srig);


/* sdump.c */

#define MAXSDUMP 2

struct sdump{
	int fd[MAXSDUMP];
	char rdbuf[MAXRDBUF];
    int colors[MAXSDUMP];
};

struct sdump *init_sdump(void);
int sdump_open(int idx, char *filename);
void sdump_read_handler(void *);

#if 0
/* mike.c */

struct mike{
    pthread_t thread;
    unsigned long ssrc;  /* RTP synchronisation source identifier */
    ZMUTEX_DEFINE(ssrc);
    int sending;
    int keepalive;
};

struct mike *init_mike(void);
int free_mike(struct mike *mike);
void *mike_thread_func(void *arg);

int mike_set(int compress, int flags, char *nat, int asquelch, int asqdelay);

/* speaker.c */

extern struct speaker *speaker;
struct speaker{
    pthread_t thread;
    unsigned long ssrc;  /* RTP synchronisation source identifier */
    ZMUTEX_DEFINE(ssrc);
};

struct speaker *init_speaker(void);
int free_speaker(struct speaker *speaker);
void *speaker_thread_func(void *arg);

int speaker_set(int flags, int ajitter, int ajitteridlet);

#endif


/* voip.c */


struct packet{
    uint32 n_seq;       // network order
    char buf[1];
}__attribute__((packed));

struct jitterbuf{
    struct jitterbuf *next;
    struct jitterbuf *prev;
    struct packet *packet;
    uint32 seq;         // host order
};

struct voip{
    int udpsock; 

    pthread_t mike_thread;
    uint32 mike_seq;
    int mike_sending;
    int mike_keepalive;
    
    pthread_t speaker_thread;
    struct sound *sound;
    int sound_avail_us, sound_avail_limit;

    int mike_sin, speaker_sin;
    int cw_grab, cw_active;
    int cw_vox_timer;
};

extern struct voip *voip;

struct voip *init_voip(void);
int free_voip(struct voip *voip);

void *mike_thread_func(void *arg);
void *speaker_thread_func(void *arg);

void gen_sin(void *buf, struct sound_stream *stream, int hz);

#if 0

/* sound.c */
int init_sound(int iomode);
void free_sound(void);
extern void (*sound_term)(int iomode);
extern void (*sound_open)(int *audio_io, int *audio_ctl);
extern void (*sound_play)(int len, unsigned char *buf);
extern void (*sound_playvol)(int value);
extern void (*sound_recgain)(int value);
extern void (*sound_dest)(int where);
extern int  (*sound_grab)(char *buf, int len);
extern void (*sound_flush)(void);
extern snd_pcm_sframes_t (*sound_get_delay)(void);

audio_encoding_t sound_encoding;

/* audio_oss.c */

int oss_init(int iomode);
void oss_term(int iomode);
void oss_open(int *audio_io, int *audio_ctl);
void oss_play(int len, unsigned char *buf);
void oss_playvol(int value);
void oss_recgain(int value);
void oss_dest(int where);
int oss_grab(char *buf, int len);
void oss_flush(void);
snd_pcm_sframes_t oss_get_delay(void);

/* audio_alsa.c */
#ifdef HAVE_ALSA
int alsa_init(int iomode);
void alsa_term(int iomode);
void alsa_open(int *audio_io, int *audio_ctl);
void alsa_play(int len, unsigned char *buf);
void alsa_playvol(int value);
void alsa_recgain(int value);
void alsa_dest(int where);
int alsa_grab(char *buf, int len);
void alsa_flush(void);
snd_pcm_sframes_t alsa_get_delay(void);
#endif

#endif 

/* zrandom */
//int zrandom(void);

/* help */
extern char *help;


/* ihu.c */

struct ihu{
    pid_t pid;
    char *cmdline;
};

extern struct ihu *ihu;

struct ihu *init_ihu(void);
void free_ihu(struct ihu *ihu);
void ihu_sigchld(void *xxx);
int ihu_start(struct ihu *ihu);
int ihu_stop(struct ihu *ihu);


