/*
    ihu - ihu control
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct ihu *ihu;

struct ihu *init_ihu(void){
    char *sound_type, *cmdline;

    
    sound_type = rc_str("sound_type", NULL); /* Sound type oss, alsa or ihu */
    if (!sound_type){
        dbg("no ihu\n");
        return NULL;
    }

    if (strcmp(sound_type, "ihu") != 0) return 0;
    dbg("initializing ihu\n");

    struct ihu *ihu = mem_alloc(sizeof(struct ihu));
    memset(ihu, 0, sizeof(ihu));

    if (opt_s)
        cmdline = rc_str("sihu_cmdline", "ihu --nogui --wait"); /* command to start ihu server */
    else
        cmdline = rc_str("cihu_cmdline", "ihu --nogui --call karotka"); /* command to start ihu client */

    ihu->cmdline = strdup(cmdline); 

    install_signal_handler(SIGCHLD, ihu_sigchld, ihu, 1); 
    
    cmdline = rc_str("ihu_kill", "killall ihu"); /* command to kill all ihu processes on startup */
    system("cmdline");

    ihu_start(ihu);
    return ihu;
}

void free_ihu(struct ihu *ihu){
    ihu_stop(ihu);

    install_signal_handler(SIGCHLD, NULL, ihu, 0); 
    free(ihu->cmdline);
    mem_free(ihu);
}

void ihu_sigchld(void *xxx){
    int ret;

    ret = wait(NULL);
    dbg("ihu_sigchld pid=%d\n", ret);
    if (ihu && ret == ihu->pid) ihu->pid = -1;

}

int ihu_start(struct ihu *ihu){
    pid_t pid;

    pid = fork();
    if (pid < 0){
        errormsg("ihu_start: can't fork");
        return pid;
    }           
    if (pid == 0){ // child
        int i;
        usleep(100000);
        for (i = 3; i < 1024; i++){ 
            close(i);
        } 

        execlp("/bin/sh", "sh", "-c", ihu->cmdline, NULL);
        exit(0);
    }
    // parent
    ihu->pid = pid;
    dbg("ihu_start pid=%d\n", pid);
    return 0;
}

int ihu_stop(struct ihu *ihu){
    int i;
    pid_t pid;

    dbg("ihu_stop pid=%d\n", ihu->pid);
    pid = ihu->pid;
    if (ihu->pid < 0) return ihu->pid;

    kill(pid, SIGTERM);
    for (i=0; i<10; i++){
        if (ihu->pid < 0) break;
        sleep(100);
        if (ihu->pid < 0) break;
        dbg("kill with SIGKILL\n");
        kill(pid, SIGKILL);
    }
    dbg("ihu stopped\n");
    return 0;
}
