/*
    iprig - ham radio via internet
    Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

int opt_c=0, opt_d=0, opt_e=0, opt_h=0, opt_i=0, opt_m=0, opt_p=0, opt_s=0, opt_t, opt_v=0, opt_u=0;
char *opt_hostname=NULL;
char *opt_test=NULL;
static int ac;
static char **av;

void usage(){
    printf(
"\n"            
" iprig - Remote radio via IP networks\n"
" Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz\n"
"\n"
" Command line arguments:\n"
" -d debug mode\n"
" -h this help plus configuration file format\n"
" -i informations about HW etc\n"
" -s server mode\n"
" -t test (internal)\n"
" -u dump mode (see sdump.c)\n"
" -v version\n"
"\n");
}
         
char *parse_options(int argc, char **argv){
    int c;

    while (1){
        c = getopt(argc, argv, "cdhimpst:uv");
        if (c == -1) break;

        switch (c){
            case 'c':
                opt_c++;
                break;

            case 'd':
                opt_d++;
                break;

            case 'e':
                opt_e++;
                break;

            case 'h':     
                opt_h++;
                break;

            case 'i':
                opt_i++;
                break;

            case 'm':
                opt_m++;
                break;
           
            case 'p':
                opt_p++;
                break;

            case 's':
                opt_s++;
                break;
           
            case 't':
                opt_t++;
                break;
           
            case 'v':
                opt_v++;
                break;

            case 'u':
                opt_u++;
                break;

            default:
                printf ("?? getopt returned character code 0%o ??\n", c);
        }
    }
    if (optind < argc){
        opt_hostname = argv[optind++];
    }

    if (optind < argc){
        printf ("non-option ARGV-elements: ");
        while (optind < argc) printf ("%s ", argv[optind++]);
        printf ("\n");
    }

    return NULL;
}
     

void sigterm(void *xxx){
    dbg("sigterm\n");
    terminate = 1;
}

void sigint(void *xxx){
    dbg("sigint\n");
    terminate = 1;
}

/*void testik(void){
    struct sound_stream *stream;
    
    stream = z_new0(struct sound_stream);
    stream->sndfmt = z_new0(struct sound_fmt);    
    stream->datafmt = z_new0(struct sound_fmt);    

        stream->datafmt->fmt = SOUND_FMT_S16; 
        stream->datafmt->channels = 1; 
        stream->datafmt->rate = 8000; 
        stream->datafmt->per_time_us = 25000;
        stream->datafmt->buf_time_us = 75000;
        sound_stream_open(stream, "hw:0,0");

} */

void init(){
//    char *c;
//        init_debug(2, NULL); //TODO
//    testik    
    
    parse_options(ac, av);
    
    if (opt_v){
        printf ("iprig 0.1\nCopyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>\n");
        exit(0);
    }
    
    if (opt_h){
        usage();
		printf("\nConfiguration file /etc/iprig or ~/.iprig has format:\n"
				"key value\n\n"
				"Keys are:\n"
				"%s", help);
        exit(0);
    }

    read_rc_files("iprig");
    if (opt_i){
        rc_info();
        usb_info();
		sound_info();
        video_info();
        exit(0);
    }
    
    if (opt_u){
        init_sdump();
        return;
    }
    
    if (opt_d){
        dbg("debug level %d\n", opt_d);
        init_debug(2, NULL);
    }

    srand(time(NULL));
    install_signal_handler(SIGTERM, sigterm, NULL, 0);
//    install_signal_handler(SIGINT, sigint, NULL, 0);   
    init_tcp();
//    oss_mixer_set_from_rc(&cfg);
    voip = init_voip();
    if (!voip) exit(0);
    ihu = init_ihu();
    
    if (opt_s){
        /* server */
        init_srig();
    }
    
    if (!opt_s || opt_m){
        /* client */
        if (voip) voip->mike_sending = 0;
		init_crig();
            
        if (opt_m) {
            opt_hostname="127.0.0.1";
        }else{
            if (!opt_hostname){
                opt_hostname = rc_getopt("server"); /* Hostname/IP of server to connect to */
                if (!opt_hostname){
                    usage();
                    exit(0);
                }
            }
        }
		init_term();
        tcp_connect(opt_hostname, tcp_port);
    }
}

void terminate_all_subsystems(){
	dbg("cleanup\n");
	if (voip) free_voip(voip);
    if (ihu) free_ihu(ihu);
	if (!opt_s){
		free_term();
	}
    free_debug();
}

int main(int argc, char *argv[])
{
    ac = argc;
    av = (char **)argv;

    
    select_loop(init);
    terminate_all_subsystems();
    
    return 0;
}

