/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

int init_sftdi(struct srig *srig){
    int ftdi_vid, ftdi_pid;
    int ret, f, ftdi_baudrate, i;
    char *ftdi_serial;
    int verbose = 1;
    char *ftdi_prot;
    int run_ic706p = 0;
//    char errbuf[1024];
//  char *device;
    
    dbg("init_sftdi\n");

    srig->srig_free   = free_sftdi;
    srig->srig_reset  = sftdi_reset;
    srig->srig_gpo    = sftdi_gpo;
    srig->srig_set_output = sftdi_set_output;
    srig->srig_force_send_gpi    = sftdi_force_send_gpi;

    if (pipe(srig->ftdi_pipe)) internal("Can't create sftdi pipe");
    srig->ftdi_zlb = zlb_new(256);
    
    for (i=0; i < MAXGPO; i++) srig->gpo[i] = 1;

    for (f = 0; f < MAXFTDI; f++){
        struct ftdi_context *ftdi;

        ftdi_vid = rc_intn("sftdi_vid", f, 0); /* USB vendor ID */
        ftdi_pid = rc_intn("sftdi_pid",  f,0); /* USB product ID */
        if (!ftdi_vid || !ftdi_pid) continue;

        ftdi_serial = rc_strn("sftdi_serial", f, ""); /* FTDI serial number */
        ftdi_baudrate = rc_intn("sftdi_baudrate", f, 300); /* FTDI baud rate (300-3000000) */
        srig->ftdi_bitmode[f] = rc_intn("sftdi_bitmode", f, 1); /* FTDI bitmode (0=serial, 1=bitbang etc) */
        ftdi_prot = rc_strn("sftdi_prot", f, ""); /* FTDI serial protocol (ic706p) */
        dbg("ftdi_prot='%s'\n", ftdi_prot);
        if (ftdi_prot && *ftdi_prot && strcmp(ftdi_prot, "ic706p")==0) {
            dbg("enabling PROT_IC706P\n");
            srig->ftdi_prot[f] = PROT_IC706P;
        }
        srig->ftdi_bitmask[f] = 0x00;
        srig->ftdi_gpo[f] = 0xff;
       // ftdi_dummy = rc_int("ftdi_dummy", 0); /* Dummy mode without FTDI device */

        if (!ftdi_vid) {
            error("Refuse to open VID=%04x, probably incorrect configuration.\n", ftdi_vid);
            return 1;
        }

        dbg("ftdi_open %04x:%04x serial='%s' server=%d\n", ftdi_vid, ftdi_pid, ftdi_serial, opt_s);
        
//        if (ftdi != NULL) return 0;
        pthread_mutex_init(&srig->ftdi_gpi_mutex[f], NULL);

        srig->ftdi[f] = ftdi = ftdi_new();
        if (!ftdi){
            if (verbose) error("Can't init ftdi library for ftdi");
            return 1;
        }
        if (ftdi_serial && *ftdi_serial)
            ret=ftdi_usb_open_desc(ftdi, ftdi_vid, ftdi_pid, NULL, ftdi_serial);
        else
            ret=ftdi_usb_open(ftdi, ftdi_vid, ftdi_pid);
        
        if (ret){
            if (verbose) {
                error("Can't open ftdi device %04x:%04x, error=%d %s\n", ftdi_vid, ftdi_pid, ret, ftdi->error_str);
                if (ret==-8) error("Maybe run program as root\n");
            }
            free_sftdi(srig);
            return ret;
        }
        
        ret=ftdi_set_baudrate(ftdi, ftdi_baudrate);
        if (ret){
            if (verbose) error("Can't set baudrate for ftdi, error=%d %s\n", ret, ftdi->error_str);
            free_sftdi(srig);
            return ret;
        }
        
        if (srig->ftdi_prot[f] == PROT_IC706P){
            srig->ftdi_ic706p_last[f] = 0xfd; // to allow to send keepalive
            pthread_mutex_init(&srig->ftdi_ic706p_mutex[f], NULL);
            run_ic706p++;
        }
        
    }
        
    srig_reset();

    set_handlers(srig->ftdi_pipe[0], sftdi_read_handler, NULL, NULL, srig);
    pthread_create(&srig->ftdi_async_thread, NULL, sftdi_async_thread_func, srig);

    if (run_ic706p){
        pthread_create(&srig->ftdi_ic706p_thread, NULL, sftdi_ic706p_thread_func, srig);
    }

    return 0;
}

void free_sftdi(struct srig *srig){
//    char errbuf[1024];
    int f;
    
    if (!srig) return;
    
    if (srig->ftdi_async_thread) pthread_cancel(srig->ftdi_async_thread);
    if (srig->ftdi_ic706p_thread) pthread_cancel(srig->ftdi_ic706p_thread);
    close(srig->ftdi_pipe[0]);
    close(srig->ftdi_pipe[1]);
    zlb_free(srig->ftdi_zlb);
    for (f = 0; f < MAXFTDI; f++){
        struct ftdi_context *ftdi;

        ftdi = srig->ftdi[f];
        dbg("free_sftdi %p\n", ftdi);
        if (!ftdi) continue;
        pthread_mutex_destroy(&srig->ftdi_gpi_mutex[f]);

        if (srig->ftdi_prot[f] == PROT_IC706P){
            pthread_mutex_destroy(&srig->ftdi_ic706p_mutex[f]);
        }
        ftdi_free(ftdi);
        srig->ftdi[f] = NULL;
    }    
	
	return;
}

int sftdi_reset(struct srig *srig){

    return 0; 
}


int sftdi_set_output(struct srig *srig, int gponr, int output){
    int f = gponr >> 3;
    int b = gponr & 0x07;
    struct ftdi_context *ftdi = srig->ftdi[f];
    int ret;
    
    if (gponr == IO_FLUSH){
        for (f = 0; f < MAXFTDI; f++){
            ftdi = srig->ftdi[f];
            if (!ftdi) continue;

            ret=ftdi_set_bitmode(ftdi, srig->ftdi_bitmask[f], srig->ftdi_bitmode[f]);
            dbg("ftdi_set_bitmode(%d, 0x%02x, %d)=%d\n", f, srig->ftdi_bitmask[f], srig->ftdi_bitmode[f], ret);
            if (ret){
                error("Can't enable bitbang, error=%d %s\n", ret, ftdi->error_str);
                return ret;
            }
            
            srig->ftdi_gpo[f] |= ~srig->ftdi_bitmask[f];
            ret = ftdi_write_data(ftdi, (unsigned char *)&srig->ftdi_gpo[f], 1);
            if (ret!=1) {
                error("Can't write, error=%d %s\n", ret, ftdi->error_str);
            }
        }
        return 0;
    }
    
    if (!ftdi) return -1;

    if (output) srig->ftdi_bitmask[f] |= 1 << b;
    else        srig->ftdi_bitmask[f] &= ~(1 << b);

    return 0;
}

int sftdi_gpo(struct srig *srig, int gponr, int on){
    int ret;
    int f = gponr >> 3;
    int b = gponr & 0x07;
    struct ftdi_context *ftdi = srig->ftdi[f];
    
    if (!ftdi) return -1;

    if (on) {
        srig->ftdi_gpo[f] |= 1 << b;
        srig->gpo[gponr] = 1;
    }else{
        srig->ftdi_gpo[f] &= ~(1 << b);
        srig->gpo[gponr] = 0;
    }

//    printf("ftdi_write %02x ", (unsigned char)srig->ftdi_gpo[f]);
    srig->ftdi_gpo[f] |= ~srig->ftdi_bitmask[f];
//    printf(" %02x\n", (unsigned char)srig->ftdi_gpo[f]);
//    dbg("sftdi_gpo( , %d, %d)\n", gponr, on);
    ret = ftdi_write_data(ftdi, (unsigned char *)&srig->ftdi_gpo[f], 1);
    if (ret!=1) {
        error("Can't write, error=%d %s\n", ret, ftdi->error_str);
    }
    
    return 0;
}

int sftdi_force_send_gpi(struct srig *srig){
    int f;
    struct ftdi_context *ftdi;
    
    for (f = 0; f < MAXFTDI; f++){
        ftdi = srig->ftdi[f];
        if (!ftdi) continue;
        
        pthread_mutex_lock(&srig->ftdi_gpi_mutex[f]);
        srig->ftdi_gpi[f] ^= 0xff;
        pthread_mutex_unlock(&srig->ftdi_gpi_mutex[f]);
    }
    return 0;
}


void *sftdi_async_thread_func(void *arg){
    struct srig *srig = (struct srig *)arg;
    int n, ret, i, f, b;
    struct ftdi_context *ftdi;
    char data[256];
    char s[50];

    while(1){
//        dbg("sftdi_async_thread_func\n");
        pthread_testcancel();
        n=0;
        for (f = 0; f < MAXFTDI; f++){
            ftdi = srig->ftdi[f];
            if (!ftdi) break;
            ret = ftdi_read_data(ftdi, (unsigned char *)data, sizeof(data));
//            dbg("ret=%d\n",ret);
            if (ret<=0) continue;
            n += ret;

            switch(srig->ftdi_bitmode[f]){
                case BITMODE_RESET:
                    /*dbg("rd%d(%d): ", f, ret);
                    for (i=0; i<ret; i++) {dbg("%02x ", (unsigned char)data[i]); if (i>40) break;}
                    dbg("\n");*/
                    zs_writeprintfesc(srig->ftdi_pipe[1], "%d;%h\n", f, data, ret);
                    break;
                case BITMODE_BITBANG:
                    for (i=0; i<ret; i++){
                        char xor;
                        pthread_mutex_lock(&srig->ftdi_gpi_mutex[f]);
        //                dbg("%02x\n", (unsigned char)data[i]);
                        xor = data[i] ^ srig->ftdi_gpi[f];
                        xor &= ~srig->ftdi_bitmask[f];
                        if (xor){
//                        dbg("xor=%02x bitmask=%02x\n", (unsigned char)xor, (unsigned char)srig->ftdi_bitmask[f]);
                            srig->ftdi_gpi[f] = data[i];
                            for (b = 0; b < 8; b++){
                                if (!(xor & (1<<b))) continue;
                                sprintf(s, "%d;%d;%d\n", f, b, !!(data[i] & (1 << b)));
                                write(srig->ftdi_pipe[1], s, strlen(s));
                            }
                        }
                        pthread_mutex_unlock(&srig->ftdi_gpi_mutex[f]);
                    }
                    break;
            }
        }
        if (!n) usleep(10000);
    }

    return NULL;
}

void sftdi_read_handler(void *xxx){
    int ret;
    char buf[FBUFLEN+1];
    struct zstring *zs, *zs2;
    struct srig *srig = (struct srig *)xxx;
    
    ret=read(srig->ftdi_pipe[0], buf, FBUFLEN);
//    dbg("\nsftdi_read_handler ret=%d\n", ret);
    if (ret <= 0) internal("Can't read from sftdi pipe");
    buf[ret] = '\0';
    zlb_append(srig->ftdi_zlb, buf);
//    dbg("zlb='%s'\n", srig->ftdi_zlb->str);

    while ((zs=zlb_get_line(srig->ftdi_zlb)) != NULL){
        int f, b, v;
        char *c;

        f = atoi(zs_tokenize(zs, 1));
        switch(srig->ftdi_bitmode[f]){
            case BITMODE_RESET:
                c = zs_tokenize(zs, 0);
                zs2 = zs_printfesc("SER;%d;%s", f, c);
                rel_write_all(zs2->str);
                zs_free(zs2);
                break;    
            case BITMODE_BITBANG:
                b = atoi(zs_tokenize(zs, 0));
                v = atoi(zs_tokenize(zs, 0));
        //        dbg("sftdi_read_handler(f=%d, b=%d, v=%d)\n", f, b, v);
                zs_free(zs);
                srig_async_gpi(srig, f * 8 + b, v); 
                break;
        }
    }
}

int sftdi_write(struct srig *srig, int f, char *data, int len){
    int ret;

/*    int i;
    dbg("wr%d(%d): ", f, data);
    for (i=0; i<len; i++) {dbg("%02x ", (unsigned char)data[i]); if (i>40) break;}
    dbg("\n");
*/
    if (len <= 0) return 0;
    srig->ftdi_ic706p_last[f] = data[len - 1];
    pthread_mutex_lock(&srig->ftdi_ic706p_mutex[f]);
    ret = ftdi_write_data(srig->ftdi[f], (unsigned char *)data, len);
    pthread_mutex_unlock(&srig->ftdi_ic706p_mutex[f]);
    return ret;
}

void *sftdi_ic706p_thread_func(void *arg){
    int f;
    unsigned char *keepalive = (unsigned char *)"\xfe\x0b\x00\xfd";

    while(1){
        pthread_testcancel();
        for (f = 0; f < MAXFTDI; f++){
            if (!srig->ftdi[f]) continue;
            if (srig->ftdi_prot[f] !=  PROT_IC706P) continue;

/*            if (srig->ftdi_ic706p_last[f] != (char)0xfd) 
                error("ERROR! ftdi_ic706p_last[%d] = %02x != 0xfd\n", f, (unsigned char)srig->ftdi_ic706p_last[f]);
  */
            //dbg("send keepalive f=%d\n", f);
            pthread_mutex_lock(&srig->ftdi_ic706p_mutex[f]);
            ftdi_write_data(srig->ftdi[f], keepalive, 4);
            pthread_mutex_unlock(&srig->ftdi_ic706p_mutex[f]);
        }
        usleep(100000); // 100 ms
    }
    return NULL;
}
