/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct srig *srig = NULL;

int init_srig(){
    int ret;
    char *rigtype;

   // dbg("init_srig\n");
    rigtype = rc_getopt("srig_type"); /* Type of srig (ppdev, sior, siop) */
    if (!rigtype){
        error("rig_type unspecified\n");
        return -1;
    }
	srig = mem_alloc(sizeof(struct srig));
	memset(srig, 0, sizeof(struct srig));

    if (strcasecmp(rigtype, "ppdev")==0) {
        ret = init_ppdev(srig);
        if (ret) return ret;
    }
/*TODO    if (strcasecmp(rigtype, "sior")==0) {
        ret = init_sior(srig);
        if (ret) return ret;
    } */
    if (strcasecmp(rigtype, "sftdi")==0) {
        ret = init_sftdi(srig);
        if (ret) return ret;
    }
    
    if (srig->srig_set_output){
        srig->srig_set_output(srig, IO_PWR, 1);
        srig->srig_set_output(srig, IO_PWK, 1);
        srig->srig_set_output(srig, IO_8V,  0);
        srig->srig_set_output(srig, IO_CW,  1);
        srig->srig_set_output(srig, IO_PTT, 1);
        srig->srig_set_output(srig, IO_PA,  1);
        srig->srig_set_output(srig, IO_LNA, 1);
        srig->srig_set_output(srig, 7, 1);

        srig->srig_set_output(srig, IO_FLUSH, 0);
    }

    return 0;
}

void free_srig(){
    if (srig && srig->srig_free) srig->srig_free(srig);
	mem_free(srig);
    srig = NULL;
}

char *io_name(int no){
    switch(no){
        case IO_PWR: return "PWR"; break;
        case IO_PTT: return "PTT"; break;
        case IO_8V: return "8V"; break;
        case IO_IO0: return "IO0"; break;
        case IO_PWK: return "PWK"; break;
        case IO_PA: return "PA"; break;
        case IO_LNA: return "LNA"; break;
        case IO_CW: return "CW"; break;
        case RIO_SDA: return "SDA"; break;
        case RIO_SCL: return "SCL"; break;
        default: return "???"; break;
    }
}

int srig_reset(void){
    if (!srig) return -1;

    srig_gpo(IO_PTT, 0);
    srig_gpo(IO_CW, 0);
    
    if (!srig->srig_reset) return 0;
    return srig->srig_reset(srig);
}

int srig_gpo(int gponr, int on){
	dbg("srig_gpo(%s, %d)\n", io_name(gponr), on);
	if (!srig || !srig->srig_gpo) return -1;
    switch (gponr){
        case IO_PTT:
            if (on){
				printf("TX\n");
                if (voip) voip->mike_sending = 0;
                
            }else{
				printf("RX\n");
                if (voip) voip->mike_sending = 1;
            }
            break;
    }

	srig->srig_gpo(srig, gponr, on);
    return 0;
}

int srig_force_send_gpi(){
    if (!srig) return -1;

    dbg("srig_force_send_gpi()\n");
    if (!srig->srig_force_send_gpi) return 0;
    return srig->srig_force_send_gpi(srig);
}

void srig_async_gpi(struct srig *srig, int gpinr, int on){ /* called from callback by "device driver" */
    char s[256];

    dbg("srig_async_gpi(%s, %d)\n", io_name(gpinr), on);
    switch(gpinr){
        case IO_8V:
            sprintf(s, "GPO;%d;%d",gpinr, on); 
            rel_write_all(s);
            break;
    }
    return;
}
