/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#define PROT_VER "1"

int tcpsock;
struct conn conn_list = {&conn_list, &conn_list};
struct conn *sconn = NULL;
struct conn *gconn = NULL;
int tcp_port, udp_port, rudp_port, sudp_port;

int init_tcp(){
    int on;
    char errbuf[1024];
    struct sockaddr_in sin;
    
    ZMUTEX_INIT(conn_list);
    ZMUTEX_INIT(sconn);
    tcp_port = rc_int("tcp_port", 7373); /* TCP port for communication  */
    udp_port = rc_int("udp_port", 7373); /* UDP port for voip */
    if (opt_s){
        rudp_port = rc_int("sudp_port", udp_port);
        sudp_port = rc_int("cudp_port", udp_port);
    }else{
        rudp_port = rc_int("cudp_port", udp_port);
        sudp_port = rc_int("sudp_port", udp_port);
    }

    dbg("opt_s=%d rudp_port=%d sudp_port=%d\n", opt_s, rudp_port, sudp_port);

    if (!opt_s && !opt_m) return 0;
    
    tcpsock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    /*dbg("   tcpsock=%d\n", net->tcpsock);*/
    if (tcpsock < 0){
        error("Can't create socket for tcp - %s\n", STRERR(errno));
        return -1;
    }

#ifndef __CYGWIN__    
    on=1;
    if (setsockopt(tcpsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))){
        error("Can't set SO_REUSEADDR - %s\n", STRERR(errno));
        goto x;
    }
#endif    
    
    if (fcntl(tcpsock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK - %s\n", STRERR(errno));
        goto x;
    }
        
    memset(&sin, 0, sizeof(struct sockaddr_in));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(tcp_port);
    sin.sin_addr.s_addr = INADDR_ANY;
    if (bind(tcpsock, (struct sockaddr *)&sin, sizeof(sin))){
        error("Can't bind TCP port %d - %s\n", tcp_port, STRERR(errno)); 
        goto x;
    }

    if (listen(tcpsock, 10)){
        error("Can't listen on socket %d, tcp port %d - %s\n", tcpsock, tcp_port, STRERR(errno));
        goto x;
    }

    set_handlers(tcpsock, tcp_accept_handler, NULL, tcp_exception_handler, NULL);

    
    return 0;
x:;
    close(tcpsock);
    return -1;
}

void tcp_accept_handler(void *xxx){
    int sock;
    struct sockaddr_in sin;
    socklen_t socklen;
    struct zstring *zs;
    char salt[100];
    char sinbuf[30];
    struct conn *conn;
    
    
    /*dbg("tcp_accept_handler()\n");*/

    socklen = sizeof(sin);
    sock = accept(tcpsock, (struct sockaddr *)&sin, &socklen);
    if (!socklen || sock<0) return;

    dbg("s%d: accepted %s:%d\n", sock, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

    conn = mem_alloc(sizeof(struct conn));
    bzero(conn, sizeof(struct conn));
    
    conn->sin.sin_family      = AF_INET;
    conn->sin.sin_addr.s_addr = sin.sin_addr.s_addr;
    conn->sin.sin_port        = sin.sin_port;

    memcpy(&conn->udpsin, &conn->sin, sizeof(conn->udpsin));
    conn->udpsin.sin_port = htons(sudp_port);
    
    
    dbg("conn->udpsin=%s\n", SINSTR(conn->udpsin));
    
    conn->sock = sock;
    tcp_set_state(conn, TCP_CONNECTED);
    
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, conn);
    conn->rdbuf2 = zlb_new(128);

    ZMUTEX_LOCK(conn_list);
    add_to_list(conn_list, conn);
    ZMUTEX_UNLOCK(conn_list);
    
    /* only server */
    zs = zs_concatesc("VER", PROT_VER, NULL);
    rel_write(conn, zs->str);
    zs_free(zs);
    
    sprintf(salt, "%04x%04x%04x%04x", zrandom(), zrandom(), zrandom(), zrandom());
        strcpy(salt, "salt");//fixme
    dbg("SALT='%s'\n", salt);
    zs = zs_concatesc("SALT", salt, NULL);
    rel_write(conn, zs->str);
    zs_free(zs);
    conn->salt = strdup(salt);
}

void client_authorized(struct conn *conn){
    char s[25];
    struct zstring *zs;

    conn->authorized = 1;
#if 0    
    ZMUTEX_LOCK(mike->ssrc);
    sprintf(s, "0x%08lx", mike->ssrc);
    ZMUTEX_UNLOCK(mike->ssrc);
#endif
    zs = zs_concatesc("SSRC", s, NULL);
    rel_write(conn, zs->str);
    zs_free(zs);
    rel_write(conn, "AUTHOK");

    //speaker_addest(conn);
    //mike_addest(conn);
    
    if (!gconn) {
        dbg("grabbing control\n");
        gconn = conn;
        //speaker_stop();
        if (voip) voip->mike_sending = 1;
        dbg("voip->mike_sending=%d\n", voip->mike_sending);
        send_grab(conn, 1);
        srig_force_send_gpi();
    }else{
        dbg("only rx mode\n");
        if (srig->gpo[IO_PTT]){
            /* todo restart speaker to take effect*/
        }
    }
    tcp_send_state(conn);
}

void tcp_connect(char *hostname, int port){
    int ret;
    char errbuf[1024];
    struct conn *conn;
    struct hostent *he;
        

    he = gethostbyname(hostname);
    if (!he){
        error("Can't resolve %s\n", hostname);
        return;
    }
    
    sconn = conn = mem_alloc(sizeof(struct conn));
    bzero(conn, sizeof(struct conn));

    conn->sin.sin_family      = AF_INET;
    memcpy(&conn->sin.sin_addr, he->h_addr_list[0], he->h_length);
    conn->sin.sin_port        = htons(port);
    
    memcpy(&conn->udpsin, &conn->sin, sizeof(conn->udpsin));
    conn->udpsin.sin_port = htons(sudp_port);

    conn->rdbuf2 = zlb_new(128);
    
    conn->prev = NULL;
    conn->next = NULL;
    
/*    dbg("tcp->connect %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));  */
   
    conn->sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
/*    dbg("socket returns %d \n", conn->sock);*/
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK\n");
        return;
    }
    
    ret = connect(conn->sock, (struct sockaddr *) &conn->sin, sizeof(struct sockaddr_in));
/*    dbg("connect returns %d (%d) '%s'\n", ret, errno, strerror_r(errno));*/
    if (ret==0){
        set_handlers(conn->sock, tcp_read_handler, NULL, tcp_exception_handler, conn);
        tcp_set_state(conn, TCP_CONNECTED);
        after_connect(conn);
        return;
    }
    if (ret<0){
        if (errno==EALREADY || errno==EINPROGRESS){
            tcp_set_state(conn, TCP_CONNECTING);
            set_handlers(conn->sock, NULL, tcp_connected_handler, tcp_exception_handler, conn);
            dbg("s%d: connecting to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
            return;
        }  
        error("connect error %s\n", STRERR(errno));
        tcp_disconnect(conn, "connect error");
    }
}

void tcp_connected_handler(void *xxx){
    struct conn *conn;
    int ret, err, cnt;
    socklen_t optlen;
    char errbuf[1024], s[1300];
    
    
    conn = (struct conn *)xxx;
/*    dbg("tcp_connected_handler %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    
    if (fcntl(conn->sock, F_SETFL, 0)){
        tcp_disconnect(conn, "Can't clear O_NONBLOCK");
        return;
    }

    cnt=10;
gsopt:;    
    optlen = sizeof(err);
    ret = getsockopt(conn->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &optlen);
/*    dbg("   getsockopt(SO_ERROR) returns %d err is %d\n", ret, err);  */
    if (ret!=0){
        tcp_disconnect(conn, "getsockopt returns !=0");
        return;
    }
    if (err!=0){
        sprintf(s, "getsockopt returns error %d - %s", err, STRERR(errno));
        if (errno==EINPROGRESS && --cnt>0) goto gsopt;
        tcp_disconnect(conn, s);
        return;
    }
    
    if (fcntl(conn->sock, F_SETFL, O_NONBLOCK)){
        tcp_disconnect(conn, "Can't set O_NONBLOCK");
        return;
    }
    tcp_set_state(conn, TCP_CONNECTED);
    set_handlers(conn->sock, tcp_read_handler, 
            NULL, tcp_exception_handler, conn);
    after_connect(conn);
}

void after_connect(struct conn *conn){
    struct zstring *zs;
    dbg("s%d: connected to %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    
    /* only client */
    zs = zs_concatesc("VER", PROT_VER, NULL);
    rel_write(conn, zs->str);
    zs_free(zs);
    
}

void tcp_disconnect(struct conn *conn, char *reason){
    struct conn *cc;

    if (conn->disconnecting) return;
    conn->disconnecting=1;

    dbg("s%d: disconnected %s:%d because of %s\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port), reason);
    set_handlers(conn->sock, NULL, NULL, NULL, NULL);
    if (conn->timer) kill_timer(conn->timer);
    if (conn == sconn) sconn = NULL;
    if (conn == gconn) gconn = NULL;
    ZMUTEX_LOCK(conn_list);
    if (conn->prev) del_from_list(conn);
    ZMUTEX_UNLOCK(conn_list);
    

    if (opt_s){
        //mike_rmdest(conn);
        if (!gconn){
            ZMUTEX_LOCK(conn_list);
            foreach(cc, conn_list) gconn=cc;
            ZMUTEX_UNLOCK(conn_list);
            if (gconn){
                send_grab(gconn, 1);
                dbg("new grab %d %s:%d\n", gconn->sock, inet_ntoa(gconn->sin.sin_addr), ntohs(gconn->sin.sin_port));
            }
        }
    }else{
        //mike_rmdest(conn);
        dbg("Client not connected\n");
//        terminate=1;
    }
    
    close(conn->sock);
    mem_free(conn);
}

void tcp_read_handler(void *xxx){
    struct conn *conn;            
    int ret, err;
    char errbuf[256];
    struct zstring *zs;

    conn = (struct conn*)xxx;
    
    ret=zlb_append_fd(conn->rdbuf2, conn->sock, 256);
    err=errno;
    if (ret<0){
        dbg("s%d: error reading socket - %s\n", conn->sock, STRERR(errno));
        tcp_disconnect(conn, "read returns <0");
        return;
    }
    if (ret==0){
        dbg("s%d: got disconnect\n", conn->sock);
        tcp_disconnect(conn, "got disconnect");
        return;
    }
    
    if (opt_d>=4) {                             
        dbg("s%d: rdbuf '%s'\n", conn->sock, conn->rdbuf2->str);
    }
    
    while ((zs=zlb_get_line(conn->rdbuf2)) != NULL){
        rel_read(conn, zs->str);
    }
    return;
}

void tcp_exception_handler(void *xxx){
    struct conn *conn;
    
    conn = (struct conn *)xxx;
    dbg("s%d: tcp_exception_handler %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port)); 
    
    tcp_disconnect(conn, "socket exception");
}

int tcp_send_state(struct conn *conn){
    
    return 0;
}

int conn_prod_state(struct conn *conn){
    if (!conn) return 0;
    if (conn->state == TCP_CONNECTED || conn->state == TCP_WAIT_ACK ) return 1;
    return 0;
}

void tcp_set_state(struct conn *conn, enum tcp_state state){
    
    if (opt_d>=4) dbg("s%d: tcp_set_state %s -> %s \n", conn->sock, tcp_state_s(conn->state), tcp_state_s(state));
    
    if (conn->timer) kill_timer(conn->timer);
    conn->timer = 0;
    conn->state = state;
    switch (conn->state){
        case TCP_INIT:
            break;
        case TCP_CONNECTING:
            conn->timer = install_timer(CONNECTION_TIMEOUT, kill_conn_timer, conn);
            break;
        case TCP_CONNECTED:
            conn->timer = install_timer(SEND_ACK_TIMEOUT, send_ack_timer, conn);
            break;
        case TCP_WAIT_ACK:
            conn->timer = install_timer(ACK_TIMEOUT, kill_conn_timer, conn);
            break;
        case TCP_DISCONNECTED:
            conn->timer = install_timer(REMOVE_TIMEOUT, remove_conn_timer, conn);
            break;
        case TCP_DEAD:
            break;    
    }
}

char *tcp_state_s(enum tcp_state state){
    switch (state){
        case TCP_INIT:
            return "INIT";
        case TCP_CONNECTING:
            return "CONNECTING";
        case TCP_CONNECTED:
            return "CONNECTED";
        case TCP_WAIT_ACK:
            return "WAIT_ACK";
        case TCP_DISCONNECTED:
            return "DISCONNECTED";
        case TCP_DEAD:
            return "DEAD";
    }
    return "???";
}

void send_ack_timer(void *xxx){
    struct conn *conn;
    char s[20];
    struct zstring *zs;

    conn = (struct conn *)xxx;
    if (opt_d>=3) dbg("s%d: send_ack_timer %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    conn->timer = 0;
    
    snprintf(s, 19, "%d", ++conn->relseq);
    zs = zs_concatesc("KA", s, NULL);
/*    dbg("s='%s' c='%s'\n", s, c);*/
    rel_write(conn, zs->str);
    zs_free(zs);
    tcp_set_state(conn, TCP_WAIT_ACK);
    if (voip) voip->mike_keepalive = 1;
}

void kill_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
    dbg("s%d: kill_conn_timer %s:%d\n", conn->sock, inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));
    conn->timer = 0;
    tcp_disconnect(conn, "connection or ack timeout");
}

void remove_conn_timer(void *xxx){
    struct conn *conn;

    conn = (struct conn *)xxx;
/*    dbg("remove_conn_timer %s:%d\n", inet_ntoa(conn->sin.sin_addr),
 *    ntohs(conn->sin.sin_port));*/
    conn->timer = 0;

    ZMUTEX_LOCK(conn_list);
    del_from_list(conn);
    ZMUTEX_UNLOCK(conn_list);
    free_conn(conn);
    mem_free(conn);
}

void free_conn(struct conn *conn){
    dbg("s%d: free_conn\n", conn->sock);
    set_handlers(conn->sock, NULL, NULL, NULL, NULL);
    if (conn->sock>=0) close(conn->sock);
    conn->sock=-1;
    if (conn->salt) free(conn->salt);
    if (conn->user) free(conn->user);
    tcp_set_state(conn, TCP_DEAD);
}

void tcp_write_handler(void *xxx){
    struct conn *conn;
    int written, err;
    char errbuf[1030];

    conn = (struct conn *)xxx;
/*    dbg("tcp_write_handler %s:%d\n", inet_ntoa(conn->sin.sin_addr), ntohs(conn->sin.sin_port));*/
    
    if (!conn->wrbuflen){
        set_handlers(conn->sock, tcp_read_handler, 
                NULL, tcp_exception_handler, conn);
        return;
    }
    written = write(conn->sock, conn->wrbuf, conn->wrbuflen);
    err=errno;
    if (opt_d>=3) dbg("s%d: tcp_write_handler '%s' (%d/%d) %s", conn->sock, conn->wrbuf, conn->wrbuflen, written,err);
    /*dbg("    towrite=%d  written=%d \n", towrite, written); */
    if (written<=0){
        error("error writing to socket %d - %s\n", conn->sock, STRERR(err));
        tcp_disconnect(conn, "write returns <=0");
        return;
    }
    if (conn->wrbuflen == written){
        conn->wrbuflen = 0;
        set_handlers(conn->sock, tcp_read_handler, 
                NULL, tcp_exception_handler, conn);
        return;
        
    }
    memmove(conn->wrbuf + written, conn->wrbuf, conn->wrbuflen - written); 
    conn->wrbuflen -= written;

}

/***************** "RELATION" LAYER ********************************/

void rel_write(struct conn *conn, char *s){
    int written,err, len;
    char errbuf[1030];
    
    if (!conn_prod_state(conn)){
        error("connection not i production state\n");
        return;
    }
    
    len=strlen(s);
    if (conn->wrbuflen + len + 1 > MAXWRBUF){
        error("socket %d wrbuf overflow (%d>%d)", conn->sock, conn->wrbuflen + len + 1, MAXWRBUF);
        tcp_disconnect(conn, "wrbuf overflow");
        return;
    }
    memcpy(conn->wrbuf, s, len);
    conn->wrbuf[len] = '\n';
    conn->wrbuflen += len + 1;

    written = write(conn->sock, conn->wrbuf, conn->wrbuflen);
    err=errno;
    if (opt_d>=3) {
        char s[MAXWRBUF+1];
        int i;
        
        memcpy(s, conn->wrbuf, conn->wrbuflen);
        for (i=0;i<conn->wrbuflen;i++) {
            if (!isprint(s[i])) s[i]='.';
        }
        s[i]='\0';
        
        dbg("s%d: rel_write '%s' (%d/%d/%d)\n", conn->sock, s, conn->wrbuflen, written,err);
        /*dbg("s%d: rel_write '%s' (%d/%d/%d)", conn->sock, conn->wrbuf, conn->wrbuflen, written,err);*/
    }
    if (written<0){
        if (err==EAGAIN){
            set_handlers(conn->sock, tcp_read_handler, 
                    tcp_write_handler, tcp_exception_handler, conn);
            return;
        }
        error("error writing to socket %d - %s\n", conn->sock, STRERR(err));
        tcp_disconnect(conn, "write returns <0, errno!=EAGAIN");
        return;
    }
        
    if (conn->wrbuflen == written){
        conn->wrbuflen=0;
        return;
    }
    memmove(conn->wrbuf + written, conn->wrbuf, conn->wrbuflen - written);
    conn->wrbuflen -= written;
    set_handlers(conn->sock, tcp_read_handler, 
            tcp_write_handler, tcp_exception_handler, conn);
                
}

void rel_write_all(char *s){
    struct conn *cc;
    
    ZMUTEX_LOCK(conn_list);
    foreach(cc, conn_list){
        if (conn_prod_state(cc)) {
            rel_write(cc, s);
        }

    } 
    ZMUTEX_UNLOCK(conn_list);
} 

void rel_read(struct conn *conn, char *line){
    struct zstring *zs;
    char *cmd, *arg1;
    struct zstring *zs2;

    if (strlen(line)==0) {
        error("rel_read: got empty line\n");
        return;
    }

    zs = zs_strdup(line);
    cmd = zs_tokenize(zs, 1);

    if (opt_d>=3) dbg("s%d: rel_read '%s'\n", conn->sock, line);
    
    if (strcmp(cmd, "KA")==0){    /* keepalive */
/*        dbg("   recvd KA, sending AK\n");*/
        arg1 = zs_tokenize(zs, 0);

        zs2 = zs_concatesc("AK", arg1, NULL);
        //dbg("arg1='%s' c='%s'\n", arg1, c);
        rel_write(conn, zs2->str);
        zs_free(zs2);
        goto x;
    }

    if (strcmp(cmd, "AK")==0){   /* keepalive ack */
/*        dbg("   recvd AK, setting NS_CONNECTED\n");*/
        /* todo AK number */
        tcp_set_state(conn, TCP_CONNECTED);
        
        goto x;
    }
    
  /************** app layer ********************/  
    
    if (opt_s) {
        if (opt_d==2) dbg("s%d: rel_sread '%s'\n", conn->sock, line);
        rel_sread(conn, cmd, zs);
    }else{
        if (opt_d==2) dbg("s%d: rel_cread '%s'\n", conn->sock, line);
        rel_cread(conn, cmd, zs);
    }
/*    dbg("app_read sock=%d '%s'\n", conn->sock, line);*/
    
    
x:;
    zs_free(zs);
}

void rel_sread(struct conn *conn, char *cmd, struct zstring *zs){
    char *c, *arg1, *arg2/*, *arg3*/;
    struct MD5Context md5;
    
    if (strcmp(cmd, "SER")==0){
        int f, ret;
        struct zstring *zs2;

        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for SER\n");
            return;
        }
        arg2 = zs_tokenize(zs, 0);
        if (!arg2){
            error("missing argument 2 for SER\n");
            return;
        }

        f = atoi(arg1);
        if (f >= MAXFTDI) {
            error("bad f=%d for SER\n", f);
            return;
        }
        if (srig->ftdi[f] == NULL) {
            error("srig->ftdi[%d] is not opened\n", f);
            return;
        }
        if (srig->ftdi_bitmode[f] != BITMODE_RESET) {
            error("srig->ftdi[%d] not in UART mode\n", f);
            return;
        }

//        dbg("arg2='%s'\n", arg2);
        zs2 = zs_import_hex(arg2);
        

        ret = sftdi_write(srig, f, zs2->str, zs2->size);
        if (ret != zs2->size){
            ftdierrormsg(srig->ftdi[f], "Can't write to ftdi[%d] (%d)", f, ret);
            zs_free(zs2);
            return;
        }
        zs_free(zs2);
        return;
    }

    if (strcmp(cmd, "AUTH")==0){        /* authentization from client */
        char key[100], *auth2;
        struct zstring *auth1;

        arg1 = zs_tokenize(zs, 0);        /* user */
        if (!arg1 || !*arg1){
            tcp_disconnect(conn, "auth: empty username");
            return;
        }
        if (strlen(arg1)>8) {
            tcp_disconnect(conn, "auth: too long username");
            return;
        }
        arg2 = zs_tokenize(zs, 0);        /* auth */
        if (!arg2 || strlen(arg2)!=32){
            tcp_disconnect(conn, "auth: bad auth");
            return;
        }
        strcpy(key, "user_");
        strcat(key, arg1);

//      rc_getopt("user_ok1zia"); /* MD5 hash of OK1ZIA's password */        
        c = rc_getopt(key); 
        if (!c || !*c){
            tcp_disconnect(conn, "auth: unknown user");
            return;
        }
        if (!conn->salt) {
            tcp_disconnect(conn, "auth: bad salt");
            return;
        }
        
        auth1 = zs_concatesc(arg1, c, conn->salt, NULL);
        MD5Init(&md5);
//        dbg("auth1='%s'\n", auth1->str);
        MD5Update(&md5, (unsigned char *)auth1->str, strlen(auth1->str));
        auth2 = (char *)MD5Final_str(&md5);
//        dbg("computed auth1='%s' ->  auth2='%s'\n but rcvd='%s'\n", auth1->str, auth2, arg2);
        free(auth1);

        if (strcmp(auth2, arg2)!=0){
            tcp_disconnect(conn, "auth: bad password");
        // FIXME    return;
        }
        conn->user=strdup(arg1);
        client_authorized(conn);
    }
    
    if (strcmp(cmd, "SSRC")==0){        /* synchronization source identifier */
        arg1 = zs_tokenize(zs, 0);
        conn->ssrc = strtoul(arg1, NULL, 16);
//        dbg("sread: recvd SSRC %08lx\n", conn->ssrc);
    }

    if (!conn->authorized) return;

    /*** only authorized connections */


    if (conn!=gconn) return;

    /*** only for grabbed control ***/

    if (strcmp(cmd, "GPO")==0){
        int gpo;

        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for GPO\n");
            return;
        }
        arg2 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 2 for GPO\n");
            return;
        }
        gpo=atoi(arg1);
        if (gpo<0 || gpo>MAXGPO){
            error("bad gpo %s\n", arg1);
            return;
        }
        srig->gpo[gpo]=atoi(arg2);
        srig_gpo(gpo, srig->gpo[gpo]);
    }
    
    if (strcmp(cmd, "SPKSIN")==0){
        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for SPKSIN\n");
            return;
        }
        if (voip) voip->speaker_sin = atoi(arg1);
    }
    
    if (strcmp(cmd, "MIKSIN")==0){
        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for MIKSIN\n");
            return;
        }
        if (voip) voip->mike_sin = atoi(arg1);
    }
}

void rel_cread(struct conn *conn, char *cmd, struct zstring *zs){
    char *arg1, *arg2/*, *arg3*/;
    char *user, *pass;
    struct zstring *zs2, *auth1, *auth2;
    char s[25];
    struct MD5Context md5;
    
    if (strcmp(cmd, "SER")==0){
        int f, ret;
        struct zstring *zs2;

        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for SER\n");
            return;
        }
        arg2 = zs_tokenize(zs, 0);
        if (!arg2){
            error("missing argument 2 for SER\n");
            return;
        }

        f = atoi(arg1);
        if (f >= MAXFTDI) {
            error("bad f=%d for SER\n", f);
            return;
        }
        if (crig->ftdi[f] == NULL) {
            error("crig->ftdi[%d] is not opened\n", f);
            return;
        }
        if (crig->ftdi_bitmode[f] != BITMODE_RESET) {
            error("crig->ftdi[%d] not in UART mode\n", f);
            return;
        }
        
        zs2 = zs_import_hex(arg2);


        ret = cftdi_write(crig, f, zs2->str, zs2->size);
        if (ret != zs2->size){
            ftdierrormsg(crig->ftdi[f], "Can't write to ftdi[%d] (%d)", f, ret);
            zs_free(zs2);
            return;
        }
        zs_free(zs2);
        return;
    }

    if (strcmp(cmd, "SALT")==0){        /* salt from server */
                
        arg1 = zs_tokenize(zs, 0);

        user = rc_getopt("user");       /* Client user name (callsign) */
        pass = rc_getopt("pass");       /* Client password (MD5 hash) */
        if (!user || !*user || !pass || !*pass){
            error("Unknown user or pass");
            tcp_disconnect(conn, "unknown user or pass");
            return;
        }
        //dbg("user=%s pass=%s\n", user, pass);
        auth1 = zs_concatesc(user, pass, arg1, NULL);
        MD5Init(&md5);
//        dbg("auth1='%s'\n", auth1->str);
        MD5Update(&md5, (unsigned char*)auth1->str, strlen(auth1->str));
        zs_free(auth1);

#if 0
        ZMUTEX_LOCK(mike->ssrc);
        sprintf(s, "0x%08lx", mike->ssrc);
        ZMUTEX_UNLOCK(mike->ssrc);
#endif
        zs2 = zs_concatesc("SSRC", s, NULL);
        rel_write(conn, zs2->str);
//        dbg("cread: rel_write(%s)\n", zs2->str);
        zs_free(zs2);

        auth2 = zs_concatesc("AUTH", user, MD5Final_str(&md5), NULL);
//        dbg("auth2='%s'\n", auth2->str);
        rel_write(conn, auth2->str);
        zs_free(auth2);
        return;
    }

    if (strcmp(cmd, "SSRC")==0){        /* synchronization source identifier */
        arg1 = zs_tokenize(zs, 0);
        conn->ssrc = strtoul(arg1, NULL, 16);
        return;
    }
    
    if (strcmp(cmd, "AUTHOK")==0){      /* auth success */
        if (voip) voip->mike_sending = 0;
        //speaker_start();
        crig_force_send_gpi();
        return;
    }

    if (strcmp(cmd, "GRAB")==0){        /* controlling of rig on server */
        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for GRAB\n");
            return;
        }
        crig->grab=atoi(arg1);
        crig_grab();
        return;
    }
    
    if (strcmp(cmd, "GPO")==0){
        int gpo;

        arg1 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 1 for GPO\n");
            return;
        }
        arg2 = zs_tokenize(zs, 0);
        if (!arg1){
            error("missing argument 2 for GPO\n");
            return;
        }
        gpo=atoi(arg1);
        if (gpo<0 || gpo>MAXGPO){
            error("bad gpo %s\n", arg1);
            return;
        }
        crig->gpo[gpo]=atoi(arg2);
        crig_gpo(gpo, crig->gpo[gpo]);
        return;
    }

}

void send_grab(struct conn *conn, int grab){
    char s[20];
    struct zstring *zs;

    snprintf(s, 19, "%d", grab);
    zs = zs_concatesc("GRAB", s, NULL);
    rel_write(conn, zs->str);
    zs_free(zs);
}
