/*
    iprig - ham radio via internet
    Copyright (C) 2007 Ladislav Vaiz <ok1zia@nagano.cz>
    and authors of web browser Links 0.96

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct term *term = NULL;

int init_term(void){
    char errbuf[1024];
    struct termios t;

    term = mem_alloc(sizeof(struct term));
    term->fd = 0;
    term->rdbuflen = 0;

    
	if (fcntl(term->fd, F_SETFL, O_NONBLOCK)) {
		error("error setting nonblock %s\n", STRERR(errno));
        terminate = 1;
        return -1;
    }
    

    memset(&t, 0, sizeof(struct termios));
    if (tcgetattr(term->fd, &t)) return -1;
    memcpy(&term->oldtermios, &t, sizeof(struct termios));
    cfmakeraw(&t);
    t.c_lflag |= ISIG;
#ifdef TOSTOP
    t.c_lflag |= TOSTOP;
#endif
    t.c_oflag |= OPOST;
    if (tcsetattr(term->fd, TCSANOW, &t)) return -1;


    set_handlers(term->fd, term_read_handler, NULL, term_exception_handler, NULL);
    
    return 0;
}


void free_term(void){
//    char errbuf[1024];
//  
    if (!term) return;
    if (term->fd<0) return;

	/*if (fcntl(term->fd, F_SETFL, 0)) {
		error("error setting nonblock %s\n", STRERR(errno));
        return;
    }    */
    tcsetattr(term->fd, TCSANOW, &term->oldtermios);
    term->fd = -1;
    
}

void term_read_handler(void *xxx){
    char s[1024], errbuf[1024];
    int ret;

    ret = read(term->fd, s, sizeof(s)-1);
    if (ret < 0){
        error("error reading from term - %s\n", STRERR(errno));
        terminate = 1;
        return;
    }
    if (ret == 0){
        error("got terminal disconnect\n");
        terminate = 1;
        return;
    }
    s[ret]='\0';
    
    if (opt_d>=2) dbg("term_read_handler '%s'\n", s);
    
    if (term->rdbuflen + ret >= MAXRDBUF - 1){
        error("terminal rdbuf overflow (%d>%d)\n", term->rdbuflen + ret, MAXRDBUF);
        terminate = 1;
        return;
    }

    memcpy(term->rdbuf + term->rdbuflen, s, ret+1); 
    term->rdbuflen += ret;
    
    if (opt_d>=2) dbg("term_read_handler '%s'\n", term->rdbuf);
    while(process_queue());
}


int process_queue(){    
    int el = 0, evx = -1, evy = 0;

    if (!term->rdbuflen) goto end;
    if (term->rdbuf[0] == '\033') {
        if (term->rdbuflen < 2) goto ret;
        if (term->rdbuf[1] == '[' || term->rdbuf[1] == 'O') {
            char c;
            int v;
            if (get_esc_code(term->rdbuf, term->rdbuflen, &c, &v, &el)) {
                if (term->rdbuflen < 4 || term->rdbuf[2] != '[') goto ret;
                if (term->rdbuf[2] != '[') goto ret;
                if (term->rdbuf[3] < 'A' || term->rdbuf[3] > 'L') goto ret;
                evx = KBD_F1 + term->rdbuf[3] - 'A';
                el = 4;
            }
            else switch (c) {
                case 'A': evx = KBD_UP; break;
                case 'B': evx = KBD_DOWN; break;
                case 'C': evx = KBD_RIGHT; break;
                case 'D': evx = KBD_LEFT; break;
                case 'F':
                case 'e': evx = KBD_END; break;
                case 'H':
                case 0: evx = KBD_HOME; break;
                case 'I': evx = KBD_PGUP; break;
                case 'G': evx = KBD_PGDN; break;

                case 'N': evx = KBD_F2; break;
                case 'O': evx = KBD_F3; break;
                case 'T': evx = KBD_F8; break;
                case 'U': evx = KBD_F9; break;
                case 'V': evx = KBD_F10; break;
                case 'W': evx = KBD_F11; break;
                case 'X': evx = KBD_F12; break;

                          
                case 'z': switch (v) {
                    case 247: evx = KBD_INS; break;
                    case 214: evx = KBD_HOME; break;
                    case 220: evx = KBD_END; break;
                    case 216: evx = KBD_PGUP; break;
                    case 222: evx = KBD_PGDN; break;
                    case 249: evx = KBD_DEL; break;
                    } break;
                case '~': switch (v) {
                    case 1: evx = KBD_HOME; break;
                    case 2: evx = KBD_INS; break;
                    case 3: evx = KBD_DEL; break;
                    case 4: evx = KBD_END; break;
                    case 5: evx = KBD_PGUP; break;
                    case 6: evx = KBD_PGDN; break;
                    case 7: evx = KBD_HOME; break;
                    case 8: evx = KBD_END; break;
                    case 11: evx = KBD_F1; break;
                    case 12: evx = KBD_F2; break;
                    case 13: evx = KBD_F3; break;
                    case 14: evx = KBD_F4; break;
                    case 15: evx = KBD_F5; break;
                    case 17: evx = KBD_F6; break;
                    case 18: evx = KBD_F7; break;
                    case 19: evx = KBD_F8; break;
                    case 20: evx = KBD_F9; break;
                    case 21: evx = KBD_F10; break;
                    case 23: evx = KBD_F11; break;
                    case 24: evx = KBD_F12; break;
                    } break;
                case 'P': evx = KBD_F1; break;
                case 'Q': evx = KBD_F2; break;
                case 'R': evx = KBD_F3; break;
                case 'S': evx = KBD_F4; break;
                case '[':
                    if (term->rdbuflen>=4){
                        el=4;      
                        switch(term->rdbuf[3]){
                            case 'A': evx = KBD_F1; goto upbrk;
                            case 'B': evx = KBD_F2; goto upbrk;
                            case 'C': evx = KBD_F3; goto upbrk;
                            case 'D': evx = KBD_F4; goto upbrk;
                            case 'E': evx = KBD_F5; goto upbrk;
                        }        
                    }
                upbrk:        
                break;          
            }
        } else {
            el = 2;
            if (term->rdbuf[1] >= ' ') {
                evx = term->rdbuf[1];
                evy = KBD_ALT;
                goto l2;
            }
            if (term->rdbuf[1] == '\033') {
                if (term->rdbuflen >= 3 && (term->rdbuf[2] == '[' || term->rdbuf[2] == 'O')) el = 1;
                evx = KBD_ESC;
                goto l2;
            }
        }
        goto l1;
    } 
    else if (term->rdbuf[0] == 0) {
        error("term->rdbuf[0]==0\n");
#if 0
        if (term->rdbuflen < 2) goto ret;
        if (!(evx = os2xtd[(unsigned int)term->rdbuf[1]].x)) evx = -1;
        evy = os2xtd[(unsigned int)term->rdbuf[1]].y;
        el = 2;
        /*printf("%02x - %02x %02x\n", (int)term->rdbuf[1], evx, evy);*/
        goto l1;
#endif    
    }
    el = 1;
    evx = term->rdbuf[0];
    l2:
    if (evx == 8) evx = KBD_BS;
    if (evx == 9) evx = KBD_TAB;
    if (evx == 10) evx = KBD_ENTER, evy = KBD_CTRL;
    if (evx == 13) evx = KBD_ENTER;
    if (evx == 127) evx = KBD_BS;
    if (evx < ' ') {
        evx += 'A' - 1;
        evy = KBD_CTRL;
    }
    l1:
    if (term->rdbuflen < el) {
        internal("event queue underflow");
        term->rdbuflen = el;
    }
    if (evx != -1) {
        //queue_event(itrm, (char *)&ev, sizeof(struct event));
        handle_key(evx, evy);
        memmove(term->rdbuf, term->rdbuf + el, term->rdbuflen -= el);
    } else {
        /*printf("%d %d\n", term->rdbuflen, el);fflush(stdout);*/
        memmove(term->rdbuf, term->rdbuf + el, term->rdbuflen -= el);
    }
    end:
    if (term->rdbuflen < MAXRDBUF) set_handlers(term->fd, term_read_handler, NULL, term_exception_handler, term);
    return el;
    ret:
    term->timer = install_timer(ESC_TIMEOUT, kbd_timeout, term);
    return 0;
}


void term_exception_handler(void *xxx){
    terminate = 1;
}

int get_esc_code(char *str, int len, char *code, int *num, int *el)
{
    int pos;
    *num = 0;
    for (pos = 2; pos < len; pos++) {
        if (str[pos] < '0' || str[pos] > '9' || pos > 7) {
            *el = pos + 1;
            *code = str[pos];
            return 0;
        }
        *num = *num * 10 + str[pos] - '0';
    }
    return -1;
}

void kbd_timeout(void *xxx)
{
    term->timer = -1;
    if (!term->rdbuflen) {
        error("timeout on empty queue\n");
        return;
    }
    handle_key(KBD_ESC, 0);
    if (--term->rdbuflen) memmove(term->rdbuf, term->rdbuf+1, term->rdbuflen);
    while (process_queue()) ;
}

void toggle_term_gpi(int gpinr){
    char s[50];

    crig->gpi[gpinr]=!crig->gpi[gpinr];
    dbg("GPI%d=%d\n", gpinr, crig->gpi[gpinr]);
    sprintf(s, "GPO;%d;%d", gpinr, crig->gpi[gpinr]);
    rel_write(sconn, s);
}

int handle_key(int key, int shifts){
    char s[50];
    
    if (key==KBD_TAB) dbg("\n");
    dbg("\nkey=%d shifts=%d\n", key, shifts);

    switch(key){
        case KBD_F10:
        case KBD_F12:
            terminate = 1;
            break;
        case 'h':
        case '?':
            printf("TAB, Cw, pwR, Lna, Pa, o0, o3, o4, pwK\n"); 
            break;
        case KBD_TAB:
            toggle_term_gpi(IO_PTT);
			if (crig->gpi[IO_PTT]){
				printf("TX\n");
				if (voip) voip->mike_sending = 1;
			}else{
				printf("RX\n");
				if (voip) voip->mike_sending = 0;
			}
            break;
        case 'c':
            toggle_term_gpi(IO_CW);
            break;
        case 'r':
            toggle_term_gpi(IO_PWR);
            break;
        case 'l':
            toggle_term_gpi(IO_LNA);
            break;
        case 'q':
            toggle_term_gpi(IO_PA);
            break;
        case 'k':
            toggle_term_gpi(IO_PWK);
            break;
        case 's':
            if (!voip) break;
            voip->speaker_sin = !voip->speaker_sin; 
            printf("speaker_sin=%d\n", voip->speaker_sin);
            sprintf(s, "SPKSIN;%d", voip->speaker_sin);
            rel_write(sconn, s);
            break;
        case 'S':
            if (!voip) break;
            voip->mike_sin = !voip->mike_sin; 
            printf("mike_sin=%d\n", voip->mike_sin);
            sprintf(s, "MIKSIN;%d", voip->mike_sin);
            rel_write(sconn, s);
            break;

    }
    
    return 0;
}
