/*
    iprig - ham radio via internet
    Copyright (C) 2007-2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#include <math.h>

struct voip *voip;

struct voip *init_voip(void){
	int on;
	char errbuf[1024];
	struct sockaddr_in sadrin;
    struct voip *voip;
    char *alsa_device, *alsa_play_device, *alsa_rec_device;
    char *sound_type;
    int buffer_time, period_time;
//    struct codec *codec;
    int ret;
    struct sound_stream *stream;

    sound_type = rc_str("sound_type", NULL); /* Sound type oss, alsa or ihu */
    if (!sound_type){
        dbg("no voip\n");
        return NULL;
    }
    if (strcmp(sound_type, "oss") != 0 &&
        strcmp(sound_type, "alsa") != 0){

        dbg("unknown sound_type=%s for voip\n", sound_type);
        return NULL;

    }

    voip = mem_alloc(sizeof(struct voip));
    memset(voip, 0, sizeof(struct voip));
    

    /* udp */
    
    voip->udpsock = -1;

    voip->udpsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (voip->udpsock < 0){
        error("Can't create socket for udp - %s\n", STRERR(errno));
		goto x;
	}

#ifndef __CYGWIN__    
    on=1;
    if (setsockopt(voip->udpsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))){
        error("Can't set SO_REUSEADDR - %s\n", STRERR(errno));
        goto x;
    }
#endif 
    
    if (fcntl(voip->udpsock, F_SETFL, O_NONBLOCK)){
        error("Can't set O_NONBLOCK - %s\n", STRERR(errno));
        goto x;
    }

	/* socket must bind for packet receiving, not for sending */
    dbg("bind udp %d\n", rudp_port);
	memset(&sadrin, 0, sizeof(struct sockaddr_in));
	sadrin.sin_family = AF_INET;
	sadrin.sin_port = htons(rudp_port);
	sadrin.sin_addr.s_addr = INADDR_ANY;
    if (bind(voip->udpsock, (struct sockaddr *)&sadrin, sizeof(sadrin))){
		error("Can't bind UDP port %d - %s\n", rudp_port, STRERR(errno)); 
		goto x;
	}



    /*** soundcard ***/
    buffer_time = rc_int("buffer_time", 75000); /* Sound buffer length in us (75000) */
    period_time = rc_int("period_time", 25000); /* Sound period length in us (25000) */ 

//	codec = init_codec_u8();

   if (strcmp(sound_type, "alsa") == 0){
        /* alsa */
        if (opt_s){
            alsa_device = rc_str("salsa_device", "default"); /* Alsa device (default hw:0,0 plughw:0,0 ...) */
            alsa_play_device = rc_str("salsa_play_device", alsa_device); /* Alsa device for play (look at alsa_device) */
            alsa_rec_device = rc_str("salsa_rec_device", alsa_device); /* Alsa device for record (look at alsa_device) */
        }else{
            alsa_device = rc_str("calsa_device", "default"); /* Alsa device (default hw:0,0 plughw:0,0 ...) */
            alsa_play_device = rc_str("calsa_play_device", alsa_device); /* Alsa device for play (look at alsa_device) */
            alsa_rec_device = rc_str("calsa_rec_device", alsa_device); /* Alsa device for record (look at alsa_device) */
        }

        voip->sound = init_sound(SOUND_DUPLEX);

        stream = voip->sound->play;
        stream->datafmt->fmt = SOUND_FMT_S16; 
        stream->datafmt->channels = 1; 
        stream->datafmt->rate = 8000; 
        stream->datafmt->per_time_us = period_time;
        stream->datafmt->buf_time_us = buffer_time;
        ret = sound_stream_open(stream, alsa_play_device);
        if (ret){
            error("Can't open %s for playing (%d)\n", alsa_play_device, ret);
            goto x;
        }
        dbg("play_data: %s ", alsa_play_device);
        sound_format_dump(stream, stream->datafmt);
        dbg("play_snd: %s ", alsa_play_device);
        sound_format_dump(stream, stream->sndfmt);

        stream = voip->sound->rec;
        stream->datafmt->fmt = SOUND_FMT_S16; 
        stream->datafmt->channels = 1; 
        stream->datafmt->rate = 8000; 
        stream->datafmt->per_time_us = period_time;
        stream->datafmt->buf_time_us = buffer_time;
        sound_stream_open(stream, alsa_rec_device);
        if (ret){
            error("Can't open %s for recording (%d)\n", alsa_rec_device, ret);
            goto x;
        }
        dbg("rec_data: %s ", alsa_play_device);
        sound_format_dump(stream, stream->datafmt);
        dbg("rec_snd: %s ", alsa_play_device);
        sound_format_dump(stream, stream->sndfmt);
        voip->sound_avail_us = rc_int("avail_time", 30000); /* Free time in sound buffer to write next audio data */
        voip->sound_avail_limit = voip->sound_avail_us / stream->datafmt->frame_time_us;
        dbg("avail=%d avail_limit=%d\n", voip->sound_avail_us, voip->sound_avail_limit);

    }else{
        error("no sound_type defined, running without voip subsystem\n");
        goto x;
    }

    /*** play test ***/
    if (opt_p){ 
        signed short *play_buf;

        stream = voip->sound->play;
        play_buf = mem_alloc(stream->datafmt->per_len_b);

        int i, ret;
/*        for (i=0; i<stream->datafmt->per_len_s; i++){
            play_buf[i] = 20000 * cos(i * 2*M_PI*800/8000); 
        }
        for (i=0;i<stream->datafmt->per_len_s;i++) dbg("%d ", play_buf[i]);
        dbg("\n");  */
  
        for(i=0;;){
            int avail;

            avail = sound_avail(stream);

            dbg("avail=%d (%d)\n", avail, voip->sound_avail_limit);
            if (avail < voip->sound_avail_limit){ 
                usleep(stream->datafmt->per_time_us / 2);
                continue;
            }
  
            gen_sin(play_buf, stream, 789); 
            ret = sound_playback(stream, play_buf, stream->datafmt->per_len_f);
            if (ret < 0) {
                snderrormsg(ret, "chyba");
                break;
            }
            i++;

        }
        exit(0);
    }

    /*** mike ***/

    voip->mike_sending = 0;
    voip->mike_seq = 0;
//    if (opt_m || opt_s)  //TODO
    if (pthread_create(&voip->mike_thread, NULL, mike_thread_func, voip)){
        error("Can't create mike thread");
        goto x;
    }


    /*** speaker ***/
    

//    if (opt_m || !opt_s)  //TODO
    if (pthread_create(&voip->speaker_thread, NULL, speaker_thread_func, voip)){
        error("Can't create speaker thread");
        goto x;
    }


    return voip;
x:;
    free_voip(voip);
    return NULL;
}


int free_voip(struct voip *voip){
    if (voip->speaker_thread) {
        pthread_cancel(voip->speaker_thread);
        voip->speaker_thread = (pthread_t)0;
    }
    
    if (voip->mike_thread) {
        pthread_cancel(voip->mike_thread);
        voip->mike_thread = (pthread_t)0;  // pthread_t is unsigned long int
    }
    
    if (voip->udpsock >= 0 ) {
        close(voip->udpsock);
        voip->udpsock = -1;
    }
    

    if (voip->sound){
        free_sound(voip->sound);
        voip->sound = NULL;
    }
    
    return 0;
}


void *mike_thread_func(void *arg)
{
    struct voip *voip = (struct voip *)arg;
    int rd, ret;
    struct packet *packet;
    int packetlen;
//    char errbuf[256];
    char sinbuf[30];
    struct conn *cc;
//	char *sndbuf;
    struct sound_stream *stream;
    
    dbg("mike started\n");
    stream = voip->sound->rec;
    packetlen = sizeof(struct packet);
    packetlen += stream->datafmt->per_len_b;
    packet = mem_alloc(packetlen);


    
    while(1){
        pthread_testcancel();

        rd = sound_record(stream, &packet->buf, stream->datafmt->per_len_f);
        if (rd < 0){
            error("Can't record from soundcard (%d)\n", rd);
            mem_free(packet);
            break;
        } 
        if (voip->mike_sin){
            gen_sin(&packet->buf, stream, 800);
        }

        voip->mike_seq++;
        packet->n_seq = htonl(voip->mike_seq);
        dbg("voip->mike_sending=%d   rd=%d\n", voip->mike_sending, rd);

        if (!voip->mike_sending) continue;
//        if (voip->mike_seq % 20 != 0) continue;
                
/*        dbg("send=");
        for (i=0; i<20; i++){
            dbg("%d ", packet->buf[i]);
        }       
        dbg("\n");  */
//        dbg("read=%d sending=%d seq=%d conn_list=%d sconn=%p\n", rd, voip->mike_sending, voip->mike_seq, list_empty(conn_list), sconn);
        ZMUTEX_LOCK(conn_list);
        foreach(cc, conn_list){
/*            dbg("  sendto(%s, %d) seq=%d   ", SINSTR(cc->udpsin), packetlen, voip->mike_seq);
            int i;for (i=0; i<10; i++) dbg("%d ", (unsigned char)packet->buf[i]);
            dbg("\n");
  */
            ret = sendto(voip->udpsock, packet, packetlen, 0,
                    (struct sockaddr *)&cc->udpsin, sizeof(struct sockaddr));
            if (ret < 0){
                error("Can't write to %s\n", SINSTR(cc->udpsin));
            }
        }
        ZMUTEX_UNLOCK(conn_list);

        if (sconn && !opt_m){
            ZMUTEX_LOCK(sconn);
/*            dbg("  sendto(%s, %d) seq=%d   ", SINSTR(sconn->udpsin), packetlen, voip->mike_seq);
            int i;for (i=0; i<10; i++) dbg("%d ", (unsigned char)packet->buf[i]);
            dbg("\n");
*/            
            ret = sendto(voip->udpsock, packet, packetlen, 0,
                    (struct sockaddr *)&sconn->udpsin, sizeof(struct sockaddr));
            if (ret < 0){
                error("Can't write to %s\n", SINSTR(sconn->udpsin));
            }
            ZMUTEX_UNLOCK(sconn);
        }
    }
    mem_free(packet);
    
    return NULL;
}


void *speaker_thread_func(void *arg)
{
    struct voip *voip = (struct voip *)arg;
    int rd, ret;
    struct sockaddr_in sin;
    socklen_t slen;
    char errbuf[256];
//    char sinbuf[30];
    struct timeval tv;
    fd_set readfds;
    struct jitterbuf jitterbuf = {&jitterbuf, &jitterbuf};
    int jitlen = 0;
    int avail, packetlen;
    struct sound_stream *stream;

    
    dbg("speaker started\n");
    stream = voip->sound->play;
    packetlen = sizeof(struct packet);
    packetlen += stream->datafmt->per_len_b;


    while(1){
        pthread_testcancel();

        FD_ZERO(&readfds);
        FD_SET(voip->udpsock, &readfds);
        tv.tv_sec = 0;
        tv.tv_usec = 5000;
        ret = select(voip->udpsock + 1, &readfds, NULL, NULL, &tv);
//        dbg("select=%d\n", ret);
        if (ret < 0){
            error("Can't select(%d) - %s\n", voip->udpsock, STRERR(errno));
            break;
        }

        if (FD_ISSET(voip->udpsock, &readfds)){
            struct packet *packet = NULL;
            struct jitterbuf *jb, *j;
            
            packet = mem_alloc(packetlen);
            slen = sizeof(struct sockaddr_in);
            rd = recvfrom(voip->udpsock, packet, packetlen, 0, 
                    (struct sockaddr *)&sin, &slen);
            if (rd < 0){
                errormsg("Error receiving UDP\n");
                break;
            }
        
            
            jb = mem_alloc(sizeof(struct jitterbuf));
            jb->packet = packet;
            jb->seq = ntohl(packet->n_seq);
            //add_to_list(jitterbuf, jb);
			foreach(j, jitterbuf) if (j->seq >= jb->seq) break;
			add_at_pos(j->prev, jb);

            jitlen++;

/*            error("recvfrom %s seq=%d  ", SINSTR(sin), jb->seq); 
            for (i=0; i<10; i++)dbg("%d ", (unsigned char)packet->buf[i]);
            dbg("\n");  
*/            
        }

        // timeout or newly added data
        avail = sound_avail(stream);
/*        if (avail < voip->sound->codec->sndlen_f){
            error("sound_avail failed (%d < %d)\n", avail, voip->sound->codec->sndlen_f);
            break;
        }*/

        //dbg("avail=%d (%d)\n", avail, voip->sound_avail_limit);
        if (avail < voip->sound_avail_limit) continue;
       
        if (jitlen==0){
            void *emptybuf; 
//            dbg("playing empty\n");
            emptybuf = mem_alloc(stream->datafmt->per_len_b);
            memset(emptybuf, 0, stream->datafmt->per_len_b);
            ret = sound_playback(stream, emptybuf, stream->datafmt->per_len_f);
            mem_free(emptybuf);
        }else{
            struct jitterbuf *jb;
            jb = jitterbuf.prev;
            del_from_list(jb);
            jitlen--;
//            dbg("playing seq=%d\n", jb->seq);
            if (voip->speaker_sin){
                gen_sin(&jb->packet->buf, stream, 811);
            }
            ret = sound_playback(stream, jb->packet->buf, stream->datafmt->per_len_f);
            mem_free(jb->packet);
            mem_free(jb);
        }
                
        if (ret < 0){
            error("Can't write to soundcard - %s\n", STRERR(errno));
            break;
        }
    }
    
    return NULL;
}

void gen_sin(void *buf, struct sound_stream *stream, int hz){
    int16_t *dst;
    int i;

    dst = (int16_t *)buf;
    for (i = 0; i < stream->datafmt->per_len_f; i++){
        *dst++ = 20000 * sin(stream->sin_phase * 2*M_PI*hz/8000);
        stream->sin_phase++;
    }
//    dbg("gen_sin: phase=%d\n", stream->sin_phase);

}
